# NYC School Data
# Copyright (C) 2022. Matthew X. Curinga
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU AFFERO GENERAL PUBLIC LICENSE (the "License") as
# published by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the License for more details.
#
# You should have received a copy of the License along with this program.
# If not, see <http://www.gnu.org/licenses/>.
# ==============================================================================
import os.path

import pandas as pd

from .dataloader import load
from . import config
urls = config.urls
__class_size_file = os.path.join(config.data_dir, config.urls["class_size"].filename)
__ptr_file = os.path.join(config.data_dir, config.urls["class_size"].filename_ptr)

def load_class_size():
    return load(config.urls["class_size"].filename)

def load_ptr():
    return load(config.urls["class_size"].filename_ptr)

def get_class_22(url):
    """Read class size data for 2022 from
       the DOE InfoHub Excel file format.
       This data also contains pupil teacher ratios
       which are saved in a separate file.
    """

    xls = pd.read_excel(url, sheet_name=None)
    # first sheet is k-8
    k8 = xls["K-8 Avg"]
    hs = xls["MS HS Avg"]
    ptr = xls["PTR"]
    ptr.columns = ['dbn', 'school_name', 'ptr']
    ptr["ay"] = 2022

    df = pd.concat([k8, hs], axis=0)
    df["ay"] = 2022

    df.columns = ['dbn', 'school_name', 'grade', 'program_type',
                'students_n', 'classes_n', 'avg_class_size',
                'min_class_size', 'max_class_size', 'dept', 'subject',
                'ay']

    return df, ptr


def get_class_size_year(ay, url):
    def fix_rows(row):
        if row.min_class_size == "<15":
            row.min_class_size = row.students_n
        if row.max_class_size == "<15":
            row.max_class_size = row.students_n
        if row.grade in ['K', 1, 2, 3, 4, 5,]:
            row.subject = "Elementary"
            row.dept = "Elementary"
        return row

    if (ay == 2022):
        df, ptr = get_class_22(url)

        ptr.to_csv(__ptr_file, index=False)
    else:
        df = pd.read_csv(url, dtype={'dbn': str})
        df["ay"] = ay
        df.rename(columns={
            "number_of_students": "students_n", 
            "grade_level": "grade",
            "number_of_classes":"classes_n",
            "minimum_class_size":"min_class_size",
            "maximum_class_size":"max_class_size",
            "average_class_size":"avg_class_size"
            }, inplace=True)
        
    df = df.apply(fix_rows, axis=1)
    return df


def get_class_size():
    """Get class size data from the web and cleans class 
    size data for each year that it is available
    in the `datasets`. Currently data is available for each year from
    2009-2021 excluding the 2020-2021 school year.

    Returns
    -------
    DataFrame
        a pandas DataFrame holding school demographic data for all of the schools
        in the data portal

    """
    years =[]
    for ay, url in urls["class_size"].data_urls.items():
        data = get_class_size_year(int(ay), url)
        years.append(data)
    df = pd.concat(years)
    df.to_csv(__class_size_file, index=False)
    return df






# NYC School Data
# Copyright (C) 2022-23. Matthew X. Curinga
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU AFFERO GENERAL PUBLIC LICENSE (the "License") as
# published by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the License for more details.
#
# You should have received a copy of the License along with this program.
# If not, see <http://www.gnu.org/licenses/>.
# ==============================================================================


# these are the URLs into the open data portals
# update when new datasets are released
urls = {
    "datasite": {
        "url": "https://data.mixi.nyc/",
        "desc": "online data portal for nycschools"
    },
    "docbook": {
        "url": "https://adelphi-ed-tech.github.io/nycschools/install.html"
    },
    "galaxy": {
        "url": "https://www.nycenet.edu/offices/d_chanc_oper/budget/dbor/galaxy/galaxyallocation/default.aspx",
        "desc": "Galaxy Budget Summaries",
        "url_stem": "https://www.nycenet.edu/offices/d_chanc_oper/budget/dbor/galaxy/galaxybudgetsummaryto/default.aspx?DDBSSS_INPUT=",
        "filename": "galaxy-budget.csv"
    },
    "cep": {
        "url": "https://www.iplanportal.com/",
        "desc": "Capital Expenditure Plan",
        "url_stem": "https://www.nycenet.edu/documents/oaosi/cep/2022-23/"
    },
    "school-data-archive": {
        "url": "https://drive.google.com/uc?export=download&id=1I35Wr1-UObcPm9CYSgqPUa8JOzOuAQBF",
        "filename": "nycschools-data.7z",
        "desc": "Download the archive of clean school data from public Google Drive"
    },
    "hs_admissions": {
        "url": "https://data.cityofnewyork.us/browse?q=Selection%20Criteria%20for%20High%20School%20Admissions",
        "data_urls": {
            "2018": "",
            "2019": "",
            "2020": "",
            "2021": "https://data.cityofnewyork.us/resource/9gs9-zhxw.csv?$limit=1000000"
        }
    },
    "hs_directory": {
        "url": "https://data.cityofnewyork.us/resource/97mf-9njv.csv?$limit=1000000",
        "desc": "High School Directories",
        "data_urls": {
            "2013": "https://data.cityofnewyork.us/resource/u553-m549.csv?$limit=1000000",
            "2014": "https://data.cityofnewyork.us/resource/n3p6-zve2.csv?$limit=1000000",
            "2015": "https://data.cityofnewyork.us/resource/by6m-6zpb.csv?$limit=1000000",
            "2016": "https://data.cityofnewyork.us/resource/7crd-d9xh.csv?$limit=1000000",
            "2017": "https://data.cityofnewyork.us/resource/s3k6-pzi2.csv?$limit=1000000",
            "2018": "https://data.cityofnewyork.us/resource/vw9i-7mzq.csv?$limit=1000000",
            "2019": "https://data.cityofnewyork.us/resource/uq7m-95z8.csv?$limit=1000000",
            "2020": "https://data.cityofnewyork.us/resource/23z9-6uk9.csv?$limit=1000000",
            "2021": "https://data.cityofnewyork.us/resource/8b6c-7uty.csv?$limit=1000000"
        }

    },
    "shsat_apps": {
        "url": "https://data.cityofnewyork.us/browse?q=SHSAT%20Admissions%20Test%20Offers%20By%20Sending%20School&sortBy=relevance",
        "filename": "shsat-applicants.csv",
        "desc": "SHSAT Applicants by School",
        "data_urls": {
            "2015": "https://data.cityofnewyork.us/resource/xqx4-kdvp.csv?$limit=1000000",
            "2016": "https://data.cityofnewyork.us/resource/8ws3-956v.csv?$limit=1000000",
            "2017": "https://data.cityofnewyork.us/resource/vsgi-eeb5.csv?$limit=1000000",
            "2018": "https://data.cityofnewyork.us/resource/uf53-ree9.csv?$limit=1000000",
            "2019": "https://data.cityofnewyork.us/resource/xuij-x4t4.csv?$limit=1000000",
            "2020": "https://data.cityofnewyork.us/resource/k8ah-28f4.csv?$limit=1000000"
        }
    },
    "zipcodes": {
        "filename": "zipcodes.geojson",
        "desc": "Load the zipcode boundaries from NYC Open Data Portal (via USPS)"
    },
    "demographics": {
        "url": "https://infohub.nyced.org/reports/students-and-schools/school-quality/information-and-data-overview",
        "data_urls": {
            "2022": "https://infohub.nyced.org/docs/default-source/default-document-library/demographic-snapshot-2018-19-to-2022-23-(public).xlsx",
            "2016": "https://data.cityofnewyork.us/resource/vmmu-wj3w.csv?$limit=1000000",
            "2013": "https://data.cityofnewyork.us/resource/s52a-8aq6.csv?$limit=1000000",
            "2006": "https://data.cityofnewyork.us/resource/ihfw-zy9j.csv?$limit=1000000",
        },
        "filename": "school-demographics.csv",
        "desc": "School demographic data from NYC Open Data Portal"
    },
    "class_size": {
        "url": "https://data.cityofnewyork.us/browse?q=Average%20Class%20Size%20by%20School&sortBy=relevance",
        "data_urls": {
            "2022": "https://infohub.nyced.org/docs/default-source/default-document-library/updated2023_avg_classsize_schl.xlsx",
            "2021": "https://data.cityofnewyork.us/resource/sgr7-hhwp.csv?$limit=1000000",
        },
        "filename": "school-class-size.csv",
        "filename_ptr": "ptr.csv",
        "desc": "Average class size by school and academic year and pupil teacher ratio for 2022-23"
    },
    "doe_school_wwww": {
        "url": "https://www.schools.nyc.gov/schools/",
        "desc": "this is the root URL part for DOE official school websites"
    },
    "building_footprints": {
        "url": "https://data.cityofnewyork.us/api/geospatial/nqwf-w8eh?method=export&format=GeoJSON",
        "city_footprints_feather": "nyc-footprints.feather",
        "school_footprints_file": "school-building-footprints.geojson",
        "desc": "NYC GEOJSON file with all building footprints"
    },
    "school_geo": {
        "url": "https://data.cityofnewyork.us/resource/a3nt-yts4.geojson?$limit=1000000",
        "filename": "school-zipcodes.geojson",
        "desc": "location points with zipcodes"
    },
    "school_locations": {
        "url": "https://data.cityofnewyork.us/resource/wg9x-4ke6.csv?$limit=1000000",
        "filename": "school_locations.geojson",
        "desc": "x,y geolocations and location meta data"
    },
    "district_geo": {
        "url": "https://data.cityofnewyork.us/api/geospatial/r8nu-ymqj?method=export&format=GeoJSON",
        "filename": "district-shapes.geojson",
        "desc": "shape file for school districts"
    },
    "charter_ela": {
        "url": "https://data.cityofnewyork.us/resource/sgjd-xi99.csv?$limit=1000000",
        "filename": "charter-ela.csv",
        "desc": "charter school 3-8 ELA results"
    },
    "charter_math": {
        "url": "https://data.cityofnewyork.us/resource/3xsw-bpuy.csv?$limit=1000000",
        "filename": "charter-math.csv",
        "desc": "charter school 3-8 math results"
    },
    "nyc_ela": {
        "url": "https://infohub.nyced.org/docs/default-source/default-document-library/school-ela-results-2013-2023-(public).xlsx",
        "url_2019": "https://data.cityofnewyork.us/api/views/hvdr-xc2s/files/4db8f0e7-0150-4302-bed7-529c89efa225?download=true&filename=school-ela-results-2013-2019-(public).xlsx",
        "filename": "nyc-ela.csv",
        "desc": "excel file with non-charter school ELA test results with demographic categories"
    },
    "nyc_math": {
        "url": "https://infohub.nyced.org/docs/default-source/default-document-library/school-math-results-2013-2023-(public).xlsx",
        "url_2019": "https://data.cityofnewyork.us/api/views/365g-7jtb/files/17910cb0-8a62-4037-84b5-f0b4c2b3f71f?download=true&filename=school-math-results-2013-2019-(public).xlsx",
        "filename": "nyc-math.csv",
        "desc": "excel file with non-charter school math test results with demographic categories"
    },
    "nyc_regents": {
        "url": "https://data.cityofnewyork.us/api/views/2h3w-9uj9/files/ae520e30-953f-47a0-9654-c77900232236?download=true&filename=2014-15-to-2018-19-nyc-regents-overall-and-by-category---public%20(1).xlsx",
        "filename": "nyc-regents.csv",
        "desc": "regents exam results"
    },
    "nysed_math_ela": {
        "comment": "Each test year has its own URL, so this item doesn't match others in format",
        "url": "https://data.nysed.gov/downloads.php",
        "urls": [
            "https://data.nysed.gov/files/assessment/20-21/3-8-2020-21.zip",
            "https://data.nysed.gov/files/assessment/18-19/3-8-2018-19.zip",
            "https://data.nysed.gov/files/assessment/17-18/3-8-2017-18.zip",
            "https://data.nysed.gov/files/assessment/16-17/3-8-2016-17.zip",
            "https://data.nysed.gov/files/assessment/15-16/3-8-2015-16.zip"
        ],
        "filename": "nysed-exams.feather",
        "desc": "NYS grades 3-8 ELA and Math test in a .zip archive"
    }
}



import pytest
from nycschools import shsat, config
import os

def test_load_admission_offers():
    df = shsat.load_admission_offers()


    cols = {'ay', 'dbn', 'feeder_name', 
            'hs_applicants_n', 'offers_n', 'testers_n'}
    
    assert cols.issubset(df.columns)
    assert df.ay.nunique() > 5
    assert df.dbn.nunique() > 500

def test_save_administration_offers():
    df = shsat.save_administration_offers()
    f = os.path.join(config.data_dir, config.urls["shsat_apps"].filename)
    assert os.path.exists(f), "File not found " + f




# NYC School Data
# Copyright (C) 2022. Matthew X. Curinga
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU AFFERO GENERAL PUBLIC LICENSE (the "License") as
# published by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the License for more details.
#
# You should have received a copy of the License along with this program.
# If not, see <http://www.gnu.org/licenses/>.
# ==============================================================================


import warnings
import functools


def suppress_warnings(warning_category=RuntimeWarning):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=warning_category)
                return func(*args, **kwargs)
        return wrapper
    return decorator



import pytest
from nycschools import config, geo, dataloader, schools
import warnings
import geopandas as gpd


def test_merge():
    a = schools.load_school_demographics()
    b = geo.load_school_locations()
    c = b.merge(a, on="dbn", how="inner")
    assert hasattr(c, "explore"), "The merged dataframe should be a GeoDataFrame with explore()"


def test_load_zipcodes():
    df = geo.load_zipcodes()
    a = set(schools.load_school_demographics().zip)
    a.remove(0)
    b = set(df.zip)
    missing = a.difference(b)
    assert len(missing) < 3, "Missing some zip codes we expected" + str(a.difference(b))
    if len(missing) > 0:
        print(f"Missing some zip codes from the zip code data set: {missing}""")

def test_load_school_locations():
    df = geo.load_school_locations()
    assert len(df) > 1800, "Too few schools found"
    assert len(df) < 2400, "Seems like too many schools"
    a = len(df)
    b = len(df.drop_duplicates(subset="dbn"))
    assert a == b, "Duplicate dbn values found"

    assert all(df.open_year.apply(lambda x: int(x) == x)), "open_year should be an integer"


def test_load_districts():
    df = geo.load_districts()

    assert len(df) > 32, "Not enough districts represented"
    assert len(df) < 40, "Seems like too many school districts for NYC, expcted 1-32,75, etc"
    expected_keys = ['district', 'area', 'length', 'geometry']
    for k in expected_keys:
        assert k in df, f"Missing expected key: {k}"


def test_load_school_footprints():
    print("loading school footprints")
    df = geo.load_school_footprints()
    print(df.columns)
    dbn = schools.load_school_demographics().dbn
    missing = set(dbn).difference(set(df.dbn))
    print("missing school footprints:", len(missing))
    for dbn in missing:
        print(dbn)


@pytest.mark.skip(reason="too slow for normal testing")
def test_get_school_footprints():
    df = geo.get_school_footprints()
    expected_keys = ['dbn', 'geometry']
    for k in expected_keys:
        assert k in df, f"Missing expected key: {k}"


def test_get_and_save_locations():
    df = geo.get_and_save_locations()
    assert isinstance(df, gpd.GeoDataFrame) and 'geometry' in df.columns, "expected a GeoDataFrame with geometry column"
    assert len(df) > 1800, "Too few schools found"
    assert len(df) < 2400, "Seems like too many schools"
    a = len(df)
    b = len(df.drop_duplicates(subset="dbn"))
    assert a == b, "Duplicate dbn values found"
    assert len(df[df.open_year.isna()]) == 0, "There should be no missing open_year values"
    # assert all(df.open_year.apply(lambda x: int(x) == x)), "open_year should be an integer"

def test_get_points():
    urls = config.urls
    df = geo.get_points()
    print(df.columns)
    assert len(df) > 1800, "Too few schools found"
    assert len(df) < 2400, "Seems like too many schools"
    a = len(df)
    b = len(df.drop_duplicates(subset="dbn"))
    assert a == b, "Duplicate dbn values found"
    expected_keys = ['dbn', 'zip', 'geo_district', 'district', 'x', 'y', 'geometry']
    for k in expected_keys:
        assert k in df, f"Missing expected key: {k}"

def test_get_locations():
    urls = config.urls
    df = geo.get_locations()
    assert len(df) > 1800, "Too few schools found"
    assert len(df) < 2400, "Seems like too many schools"
    a = len(df)
    b = len(df.drop_duplicates(subset="dbn"))
    assert a == b, "Duplicate dbn values found"

    assert all(df.open_year.apply(lambda x: int(x) == x)), "open_date should be an integer"

    expected_keys = [
        'dbn',
        'administrative_district_code',
        'administrative_district_name',
        'beds',
        'borough_block_lot',
        'census_tract',
        'community_district',
        'community_school_sup_name',
        'council_district',
        'fax_number',
        'fiscal_year',
        'geographical_district_code',
        'grades_final_text',
        'grades_text',
        'highschool_network',
        'highschool_network_location',
        'highschool_network_name',
        'latitude',
        'location_category_description',
        'location_code',
        'location_name',
        'location_type_description',
        'longitude',
        'managed_by_name',
        'nta',
        'nta_name',
        'open_year',
        'police_precinct',
        'primary_building_code',
        'principal_name',
        'principal_phone_number',
        'principal_title',
        'state_code',
        'status_descriptions']

    for k in expected_keys:
        assert k in df, f"Missing expected key: {k}"



import os.path
from nycschools import config, schools

def test_load_school_demographics():
    """Tests if schools can load demographics from local dataset and has reasonable values."""

    df = schools.load_school_demographics()
    assert df is not None, "Demo data frame is null"
    n = len(df)
    assert n > 28000, f"We should have about 28k rows, too few found ({n})"
    assert n < 30000, f"We should have about 28k rows, too many found ({n})"
    expected_keys = schools.demo.default_cols
    for k in expected_keys:
        assert k in df, f"Missing expected key: {k}"

    # at least all of these academic years should be in the data
    a = {2016, 2018, 2019, 2020, 2021, 2022}
    b = set(df.ay.unique())
    assert a.issubset(b), f"Missing some years we expected: {a.difference(b)}"

    a = {1, 26,  2, 28,  3,  4,  5, 13,  6,  7,  8,  9, 10, 14, 11, 12, 15,
     16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 27, 29, 30, 31, 32, 75, 79, 84}
    b = set(df.district.unique())
    assert a.issubset(b), f"Missing some districts we expected: {a.difference(b)}"
    percents = [c for c in df.columns if c.endswith("pct")] + ["eni"]
    for c in percents:
        assert df[c].min() >= 0, f"Percent less than zero found in {c}"
        assert df[c].max() <= 1, f"Percent greater than 1 find in {c}"
    counts = [c for c in df.columns if c.endswith("_n")]
    upper = df.total_enrollment.max()
    assert upper in range(5900, 6200), f"Max school size of {upper} outside of expected {range(5900, 6200)}"
    for c in counts:
        assert df[c].min() >= 0, f"Count value ({df[c].min()}) less than zero found in {c}"
        assert df[c].max() <= upper, f"Count value ({df[c].max()}) greater than max ({upper}) found in {c}"


def test_save_demographics():
    df = schools.save_demographics()
    assert df is not None, "Demo data frame is null"
    assert os.path.exists(schools.__demo_filename), "no demo file in data_dir"
    expected_keys = schools.demo.default_cols
    for k in expected_keys:
        assert k in df, f"Missing expected key: {k}"


def test_load_hs_directory():
    def check_year(ay):
        df = schools.load_hs_directory(ay)
        assert df is not None, "HS directory data frame is null"
        n = len(df)
        assert n > 400, f"We should have about 400 rows, too few found ({n})"
        assert "dbn" in df, "Missing dbn column"
    
    for ay in range(2013, 2021):
        check_year(ay)


import pytest
import numpy as np
import pandas as pd
from sklearn.metrics import mutual_info_score
import os.path
from nycschools import config
from nycschools import segregation as seg


test_M = 0.4255390
test_H = 0.4188083

def test_mutual_information():
    path = os.path.join(config.data_dir, "segdata.feather")
    df = pd.read_feather(path)

    M, H = seg.mutual_information(df, group="race", unit="school")
    assert abs(M - test_M) < 1e-6, f"Mutual information is {M}, not {test_M}"
    assert abs(H - test_H) < 1e-6, f"Theil H-index is {H}, not {test_H}"


def test_entropy():
    random_probs = np.random.rand(4)

    random_probs /= random_probs.sum()  # Normalize to sum to 1

    # Calculate the entropy of the random probability distribution
    ent = seg.entropy(random_probs)


    path = os.path.join(config.data_dir, "segdata.feather")
    df = pd.read_feather(path)
    N = df["n"].sum()
    p_u = df.groupby("school")["n"].sum() / N
    p_g = df.groupby("race")["n"].sum() / N
    group_E = 1.016071
    unit_E = 7.541018
    ge = seg.entropy(p_g)
    ue = seg.entropy(p_u)
    assert abs(ge - group_E) < 1e-4, f"Group entropy is {ge}, not {group_E}"
    assert abs(ue - unit_E) < 1e-4, f"Group entropy is {ue}, not {unit_E}"
    
def test_dissimilarity():
    test_data = pd.DataFrame(columns=["u","b","w"], data=[[1,50,10],[2,200,40],[3,10,100],[4,30,200],[5,10,150]])
    test_data = test_data.melt(id_vars="u", value_vars=["b","w"], var_name="g", value_name="n")
    D = seg.dissimilarity(test_data, "b")
    assert abs(D - 0.73) < .01



import pytest
from nycschools import budgets, config
import os, os.path

datafile = os.path.join(config.data_dir, config.urls["galaxy"].filename)
expected_cols = ['item', 'positions', 'budget', 'category', 'grade', 'service',
                 'subject', 'dbn', 'ay']

def test_open_webdriver():
    driver = budgets.open_webdriver()
    driver.get("https://www.example.com")
    html = driver.page_source
    assert "Example Domain" in html


def test_get_galaxy_summary():
    driver = budgets.open_webdriver()
    data = budgets.get_galaxy_summary("01M015", 2022, driver)
    driver.quit()
    assert set(data.columns) == set(expected_cols)
    assert len(data) > 5


@pytest.mark.skip(reason="too slow for normal testing")
def test_get_galaxy_budgets():
    if os.path.exists(datafile):
        os.remove(datafile)
    data, not_found = budgets.get_galaxy_budgets()
    assert set(data.columns) == set(expected_cols)
    assert len(data) > 100_000
    assert len(not_found) > 50
    assert os.os.path.exists(datafile)


def test_load_galaxy_budgets():
    data = budgets.load_galaxy_budgets()
    assert set(data.columns) == set(expected_cols)
    assert len(data) > 100_000
    assert len(data[data.dbn.isnull()]) == 0



# NYC School Data
# Copyright (C) 2023. Matthew X. Curinga
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU AFFERO GENERAL PUBLIC LICENSE (the "License") as
# published by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the License for more details.
#
# You should have received a copy of the License along with this program.
# If not, see <http://www.gnu.org/licenses/>.
# ==============================================================================
import os.path
import os
import pandas as pd
from bs4 import BeautifulSoup
from selenium import webdriver
from selenium.webdriver.chrome.service import Service
from selenium.common.exceptions import WebDriverException

from . import config, schools
from .dataloader import load

import dotenv
dotenv.load_dotenv()

__galaxyfile = os.path.join(config.data_dir, config.urls["galaxy"].filename)



def load_galaxy_budgets():
    """Loads the galaxy budgets from the local cache.
    
    Parameters
    ----------
    None
    
    Returns
    -------
    data : pandas.DataFrame
        A single DataFrame that combines all of the budget data scraped from the web
        for all schools in the database."""   

    return load(config.urls["galaxy"].filename)


def open_webdriver():
    """Opens a Selenium webdriver using either Firefox, Chrome, or Edge.
    
    Raises
    ------
    WebDriverException
        If no suitable webdriver can be found.

    Returns
    -------
    driver : selenium.webdriver.chrome.webdriver.WebDriver
        The Selenium webdriver with the --headless option.
    """
    # Define the possible browser driver classes
    drivers = [webdriver.Firefox, webdriver.Chrome,  webdriver.Edge]

    # Iterate through the browser driver classes and attempt to load each one
    for Driver in drivers:
        try:
            opt = None
            if Driver == webdriver.Chrome or Driver == webdriver.Edge:
                opt = webdriver.ChromeOptions()
            elif Driver == webdriver.Firefox:
                opt = webdriver.FirefoxOptions()

            opt.add_argument("--headless")
            driver = Driver(options=opt)

            return driver
        except WebDriverException as ex:
            print(ex)
            continue

    # If no drivers could be loaded, raise an error
    raise WebDriverException("""No suitable WebDriver could be found. 
Make sure that the latest Chrome or Firefox browser is installed.
Update the selenium package with:
pip install -U selenium""")




def __fix_cols(data, sections, dbn):
    """normalizes column names from scraped data."""

    col_map = {
        'title': 'item',
        'assignment': 'item',
        'organizational category': 'item',
        'total': 'item',
        'total.1': 'item',
        'grand total': 'item',
        'type of class/service': 'service'
    }


    data = data.copy()
    
    for i, section in enumerate(sections):
        data[i]["category"] = section
        data[i].columns = [c.lower() for c in data[i].columns]
        data[i].rename(columns=col_map, inplace=True)
        school_col = [c for c in data[i].columns if c.lower().startswith(dbn.lower())]
        if len(school_col) > 0:
            data[i].rename(columns={school_col[0]: 'item'}, inplace=True)
    return data


def get_galaxy_summary(dbn, ay, driver):
    """Gets the 'galaxy summary' budget for a school from the DOE website.
    
    Parameters
    ----------
    dbn : str
        The school's DBN.
    ay : int
        The school year, currently only the most recent school year (2022-2023) is available.
    driver : selenium.webdriver.chrome.webdriver.WebDriver
        The Selenium webdriver.
        
    Returns
    -------
    data : pandas.DataFrame
        A single DataFrame that combines all of the budget data scraped from the web
        for the school specified by `dbn` and `ay`."""

    url_stem = config.urls["galaxy"].url_stem
    url = f"{url_stem}{dbn[2:]}"

    driver.get(url)
    html = driver.page_source

    soup = BeautifulSoup(html, 'html.parser')
    # use the section header as row category
    sections = [section.get_text().strip() for section in soup.select('.TO_Section')]
    data = pd.read_html(html)
    data = __fix_cols(data, sections, dbn)
    data = pd.concat(data)
    data["dbn"] = dbn
    data["ay"] = ay

    return data

def get_galaxy_budgets():
    """Scrapes the 'galaxy summary' budget for all schools from the DOE website."""
    df = schools.load_school_demographics()
    dbns = df[df.district < 33].dbn.unique()
    driver = open_webdriver()
    budgets = []
    not_found = []
    for dbn in dbns:
        try:
            budgets.append(get_galaxy_summary(dbn, 2022, driver))
        except ValueError:
            not_found.append(dbn)
        except:
            print(f"Error scraping {dbn}")
            not_found.append(dbn + "*")

    data = pd.concat(budgets)
    data.item = data.item.str.lower()
    data.to_csv(__galaxyfile, index=False)

    return data, not_found


# NYC School Data
# Copyright (C) 2022. Matthew X. Curinga
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU AFFERO GENERAL PUBLIC LICENSE (the "License") as
# published by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the License for more details.
#
# You should have received a copy of the License along with this program.
# If not, see <http://www.gnu.org/licenses/>.
# ==============================================================================
import pandas as pd
import numpy as np
import wget

import os.path
import zipfile

from . import schools
from .dataloader import load
from . import config
import shutil

urls = config.urls


def load_nyc_nysed():
    """
    Load the subset set of the `load_nys_nysed` data for schools in the
    New York City Department of Education school demographics data set.
    """
    df = load_nys_nysed()
    nyc = schools.load_school_demographics()
    beds = nyc.beds.unique()
    return df[df.beds.isin(beds)]

def load_nys_nysed():
    """
    Load the grades 3-8 math and ela exam results for all schools and districts
    in New York State, in a long data format. This is the only data set that has
    demographic test results for charter schools in NYC. This data has all of
    the same columns as the NYC `load_math_ela_long`. Note that NYS does not
    report the same demographic categories as New York City. In particular, the
    two groups handle ENL students differently. NYC has 3 categories (current
    ell, ever ell, never ell) where NYS only has ELL and Non-ELL.

    The NYSED data includes categories that are not part of the NYC data such as
    homelessness, foster care, and parents in armed services.
    """

    # try to load it locally to save time
    # feather = os.path.join(config.data_dir, "nysed-exams.feather")
    # return pd.read_feather(feather)
    return load(urls["nysed_math_ela"].filename)


def load_nysed_ela_math_archives(urls=urls["nysed_math_ela"].urls):
    """
    Downloads all of the zip archives listed in the urls configuraiton. Extract
    the archives into a temp folder and read them as dataframes. Concats the
    dataframes into a single df and return.
    """
    tmp = os.path.join(config.data_dir, "tmp")
    if not os.path.exists(tmp):
        os.mkdir(tmp)

    # download and extract all of the .zip files from NYSED data portal
    for url in urls:
        download_and_extract(url, tmp)

    # delete all of the files we don't care about from the temp dir
    # and parse all of the ones we _do_ care about
    files = os.listdir(tmp)
    data = []

    for f in files:
        if f.endswith("mdb") or "RESEARCHER_FILE" not in f or f.startswith("."):
            os.remove(os.path.join(tmp, f))
        else:
            data.append(os.path.join(tmp, f))

    df = pd.concat([read_nysed_exam(d) for d in data])
    df = df.reset_index()

    # save as csv for outside/generic use
    out = os.path.join(config.data_dir, "nysed-exams.csv")
    df.to_csv(out, index=False)

    # this is a big file, so save as feather for internal use
    out = os.path.join(config.data_dir, urls["nysed_math_ela"].filename)
    df.to_feather(out)
    # delete the temp folder
    shutil.rmtree(tmp, ignore_errors=True)

    return df


def download_and_extract(url, tmp):
    """Download and extract the .zip archives from NYSED"""
    zippath = os.path.join(tmp, url.split("/")[-1])
    wget.download(url, out=zippath)
    zip = zipfile.ZipFile(zippath)
    zip.extractall(path=tmp)



def fix_cols(df):
    """
    Fixes the columns in the dataframe so that each NYSED year is consistent
    with the other years' columns and column name/formats match other  NYC DOE
    columns in the data portal.
    """
    nysed = df.copy()

    # make sure capitalization is consistent
    nysed.columns = [c.upper() for c in nysed.columns]

    col_map = {
        'NAME': 'school_name',
        "BEDSCODE":"beds",
        "SUBGROUP_NAME":"category",
        "ITEM_SUBJECT_AREA":"exam",
        "ITEM_DESC":"grade",
        "SY_END_DATE":"test_year",
        "TOTAL_TESTED":"number_tested",
        'TOTAL_ENROLLED': 'total_enrollment',
        'TOTAL_NOT_TESTED': 'number_not_tested',
        'L1_COUNT':"level_1_n",
        'L1_PCT':"level_1_pct",
        'L2_COUNT':"level_2_n",
        'L2_PCT':"level_2_pct",
        'L3_COUNT':"level_3_n",
        'L3_PCT':"level_3_pct",
        'L4_COUNT':"level_4_n",
        'L4_PCT':"level_4_pct",
        'L3-L4_PCT':"level_3_4_pct",
        'MEAN_SCALE_SCORE':"mean_scale_score",
        'SUBGROUP_CODE': 'subgroup_code'
    }


    # rename columns to match NYC cols
    def map_col(col):
        if col in col_map:
            return col_map[col]
        else:
            return col
    nysed = nysed.rename(columns=map_col)

    if "total_enrollment" not in nysed:
        nysed["total_enrollment"] = 0
    if "number_not_tested" not in nysed:
        nysed["number_not_tested"] = 0

    return nysed[list(col_map.values())]



def fix_data(df):
    """
    Cleans data in the dataframe so that row-level data is
    consistent across test years and matches NYC math/ela data sets:
    - student categories are consistent
    - counts and percents are consistent
    - exam category is consistent
    """

    nysed = df.copy()
    try: # this for excel
        nysed.test_year = nysed.test_year.apply(lambda x: x.date().year)
    except: # or this for csv
        nysed.test_year = nysed.test_year.apply(lambda x: int(x.split("/")[-1]))

    cat_map = {
        'English Language Learner': "Current ELL",
        'Asian or Native Hawaiian/Other Pacific Islander': 'Asian',
        'Hispanic or Latino': 'Hispanic',
        'Black or African American': 'Black',
        'Not Economically Disadvantaged': 'Not Econ Disadv',
        'Economically Disadvantaged': 'Econ Disadv',
        'Students with Disabilities': 'SWD',
        'General Education Students': 'Not SWD',
        'Non-English Language Learner': 'Never ELL'
    }

    def map_cats(cat):
        if cat in cat_map:
            return cat_map[cat]
        return cat

    # rename the student demo categories to match NYC
    nysed.category = nysed.category.apply(map_cats)

    # rename the exams to match NYC
    nysed.exam = nysed.exam.map({"ELA":"ela","Mathematics":"math"})


    # make counts into integers
    count_cols = [c for c in nysed.columns if c.endswith("_n")]
    # if we have these cols, add them too
    for c in ["number_tested", "total_enrollment", "number_not_tested", "test_year"]:
        if c in nysed:
            count_cols.append(c)
    for col in count_cols:

        nysed[col] = pd.to_numeric(nysed[col], errors='coerce')
        nysed[col] = nysed[col].fillna(0)
        nysed[col] = nysed[col].astype(int)

    nysed["ay"] = nysed.test_year - 1
    nysed["level_3_4_n"] = nysed.level_3_n + nysed.level_4_n

    # parse the grade leve string to just the grade int
    nysed.grade = nysed.grade.apply(lambda x: int(x[6]))


    # convert percents from 0-100 to 0-1 to match other data
    def fix_pct(x):
        if not x or x == "" or x =="-" or x == 0:
            return 0

        if hasattr(x, "endswith") and x.endswith("%"):
            x = x[:-1]

        try:
            x = float(x)
            if 0 <= x <= 1:
                return x
            x /= 100
            return x
        except:
            print("couldn't fix", x)
            return x

    pct_cols = [c for c in nysed.columns if c.endswith("pct")]
    for c in pct_cols:
        nysed[c] = nysed[c].apply(fix_pct)
        # nysed[c] = nysed[c].astype("float")

    nysed["mean_scale_score"] = pd.to_numeric(nysed["mean_scale_score"], downcast='integer', errors='coerce')

    return nysed


def read_nysed_exam(filename):
    """
    Read the Excel file that has all test scores for all schools and districts
    in New York State.
    """

    try:
        nysed = pd.read_csv(filename)
    except:
        nysed = pd.read_excel(filename)

    nysed = fix_cols(nysed)
    nysed = fix_data(nysed)

    # all_grades = calc_all_grades(nysed.sample(500))
    # nysed = nysed[cols]
    # all_grades = all_grades[cols]
    # nysed = nysed.append(all_grades, ignore_index=True)

    return nysed



def map_nysed_nyc(nysed):
    dbn_list = pd.read_feather("dbn-beds-list.feather", columns=["dbn","beds"])
    dbn_list[dbn_list.beds.isin(nysed.beds)].drop_duplicates()
    cols = list(nysed.columns)
    # add dbn to nysed
    combined = nysed.merge(dbn_list, on="beds", how="inner")
    cols = ["dbn"] + cols
    combined = combined[cols]
    combined.to_feather("nyc-nysed-exams.feather")

    return combined



def calc_all_grades(nysed):
    """
    NYSED data doesn't include All Grades like NYC schools, so this function
    adds an "All Grades" category with aggregate data for each school.
    """

    # add all students category to nysed
    def all_grades_for_school(row):
        """For each school, calculate the aggregate result for all grades"""
        qry = f"beds == '{row.beds}' and ay == {row.ay} and category == '{row.category}' and exam=='{row.exam}'"
        data = nysed.query(qry)

        mean_scale_score = np.average(data.mean_scale_score, weights=data.total_enrollment)
        num_tested = data.number_tested.sum()
        levels = ["1","2","3","4","3_4"]
        if num_tested == 0:
            levels_n = [0] * len(levels)
            levels_pct = [0] * len(levels)
        else:
            levels_n = [data[f"level_{i}_n"].sum() for i in levels]
            levels_pct = [n / num_tested for n in levels_n ]
        result = {
              'ay': row.ay,
              'exam': row.exam,
              'grade': "All Grades",
              'category': row.category,
              'test_year': data.test_year.max(),
              'mean_scale_score': mean_scale_score,
              'number_tested': num_tested,
              'number_not_tested': data.number_not_tested.sum(),
              'level_1_n': levels_n[0],
              'level_1_pct': levels_pct[0],
              'level_2_n': levels_n[1],
              'level_2_pct': levels_pct[1],
              'level_3_n': levels_n[2],
              'level_3_pct': levels_pct[2],
              'level_4_n': levels_n[3],
              'level_4_pct': levels_pct[3],
              'level_3_4_n': levels_n[4],
              'level_3_4_pct': levels_pct[2],
              'beds': data.beds.max()
        }
        rows.append(result)

    rows = []
    cols = ["beds","ay","exam","category","test_year"]
    t = nysed[cols]

    t = t.drop_duplicates()
    t.apply(all_grades_for_school, axis=1)

    return pd.DataFrame(rows)



# NYC School Data
# Copyright (C) 2022-2023. Matthew X. Curinga
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ==============================================================================
import os
import warnings
import os.path
from types import SimpleNamespace
from .datasets import urls as __urls


def get_version():
    """Returns the version of the nycschools package"""
    return "0.1.0"

def __read_urls():
    """Reads the urls from the datasets module"""
    urls = {}
    for k,v in __urls.items():
        urls[k] = SimpleNamespace(**v)
    return urls


def get_config():
    """Initialize the configuration settings.

Parameters
----------
None

Returns
-------
SimpleNamespace : 
    A namespace object with the following attributes:
    - data_dir : str
        The path to the data directory.
    - urls : dict
        A dictionary of URLs to download data if the local cache should be re-built.

Notes
-----
The location for local data files is determined by first looking for an environment variable called `NYC_SCHOOLS_DATA_DIR`. If this environment variable is not set, the data files are stored in a directory called `school-data` in the current directory. If this directory does not exist, it will be created.

To see and change these settings for your installation, run `python -m nycschools.dataloader`.
"""

    env_dir = os.environ.get("NYC_SCHOOLS_DATA_DIR", None)
    pwd = os.getcwd()
    local = os.path.join(pwd, "school-data")
    config = {}
    if env_dir:
        config["data_dir"] = env_dir
    else:
        config["data_dir"] = local
    
    if not os.path.exists(config["data_dir"]):
        no_data = f"""The data directory {config["data_dir"]} does not exist.
Run python -m nycschools.dataloader to download the data.
Please visit documentation on how to download the data at:
https://adelphi-ed-tech.github.io/nycschools/install.html

If you have downloaded the data and are seeing this warning, you may
need to set the environment variable NYC_SCHOOLS_DATA_DIR to the
path to the data directory.
"""

        warnings.warn(no_data)

    config["urls"] = __read_urls()

    return SimpleNamespace(**config)


config = get_config()



import pytest
from nycschools import config, exams, schools


def test_load_ela_excel():
    df = exams.load_ela_excel()
    check_state_test(df)


def test_load_math_excel():
    df = exams.load_math_excel()
    check_state_test(df)

def test_load_charter_ela():
    df = exams.load_charter_ela()
    check_state_test(df)

def test_load_charter_math():
    df = exams.load_charter_math()
    check_state_test(df)

def test_load_math():
    df = exams.load_math()
    check_state_test(df)
    cats = {'All Students', 'Not SWD', 'SWD', 'Asian', 'Black', 'Hispanic',
           'White', 'Female', 'Male', 'Econ Disadv', 'Not Econ Disadv',
           'Current ELL', 'Ever ELL', 'Never ELL'}
    for c in cats:
        assert df[df.category == c].number_tested.max() > 200, "not enough students tested in category: " + c

def test_load_ela():
    df = exams.load_ela()
    check_state_test(df)
    cats = {'All Students', 'Not SWD', 'SWD', 'Asian', 'Black', 'Hispanic',
           'White', 'Female', 'Male', 'Econ Disadv', 'Not Econ Disadv',
           'Current ELL', 'Ever ELL', 'Never ELL'}
    for c in cats:
        assert df[df.category == c].number_tested.max() > 200, "not enough students tested in category: " + c

def test_load_math_ela_long():
    df = exams.load_math_ela_long()
    assert "exam" in df.columns, "missing exam column from long data"


def test_load_math_ela_wide():
    df = exams.load_math_ela_wide()
    cols = df.columns
    math_cols = [c for c in cols if c.endswith("_math")]
    ela_cols = [c for c in cols if c.endswith("_ela")]
    assert len(math_cols) > 0
    assert len(ela_cols) > 0
    assert len(ela_cols) == len(math_cols)


def check_regents(df):
    expected_cols = {'dbn', 'school_type', 'school_level', 'regents_exam', 'year',
                     'category', 'number_tested', 'mean_score', 'below_65_n', 'below_65_pct',
                     'above_64_n', 'above_64_pct', 'above_79_n', 'above_79_pct',
                     'college_ready_n', 'college_ready_pct', 'test_year', 'ay'}
    assert df is not None, "data frame is null"
    assert set(df.columns) == expected_cols, "missing expected columns"
    assert len(df.regents_exam.unique()) > 8, "not enough regents exams"

def test_load_regents_excel():
    df = exams.load_regents_excel()
    check_regents(df)


def test_load_regents():
    df = exams.load_regents()
    check_regents(df)





# ---------------------------------------------------------------------------
def check_state_test(df):
    assert df is not None, "data frame is null"
    n = len(df)
    assert n > 20, f"Only ({n}) records found"
    expected_keys = exams.numeric_cols
    for k in expected_keys:
        assert k in df, f"Missing expected key: {k}"

    percents = [c for c in df.columns if c.endswith("pct")]
    for c in percents:
        assert df[c].min() >= 0, f"Percent less than zero found in {c}"
        assert df[c].max() <= 1, f"Percent greater than 1 find in {c}"
    counts = [c for c in df.columns if c.endswith("_n")]
    upper = df.number_tested.max()
    upper_range = range(int(upper * .8), int(upper *1.2))
    assert upper in upper_range, f"Max school size of {upper} outside of expected {upper_range}"
    for c in counts:
        assert df[c].max() <= upper, f"Count value ({df[c].max()}) greater than max ({upper}) found in {c}"



# NYC School Data
# Copyright (C) 2022-23. Matthew X. Curinga
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU AFFERO GENERAL PUBLIC LICENSE (the "License") as
# published by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the License for more details.
#
# You should have received a copy of the License along with this program.
# If not, see <http://www.gnu.org/licenses/>.
# ==============================================================================
"""
This package groups together several functions for
analyzing segregation within groups. It is tested
in reference to segregation patterns in NYC Public Schools.

References
-----------

Allen, R., & Vignoles, A. (2007). What should an index of school segregation measure? _Oxford Review of Education_, _33_(5), 643–668. https://doi.org/10.1080/03054980701366306

Elbers, B. (2021). A Method for Studying Differences in Segregation Across Time and Space. Sociological Methods & Research. https://doi.org/10.1177/0049124121986204
[code](https://elbersb.github.io/segregation/index.html)

Frankel, D. M., & Volij, O. (2011). Measuring school segregation. _Journal of Economic Theory_, _146_(1), 1–38. https://doi.org/10.1016/j.jet.2010.10.008

Reardon, S. F., & O’Sullivan, D. (2004). Measures of Spatial Segregation. _Sociological Methodology_, _34_(1), 121–162. https://doi.org/10.1111/j.0081-1750.2004.00150.x

Sakoda, J. M. (1981). A Generalized Index of Dissimilarity. _Demography_, _18_(2), 245–250. https://doi.org/10.2307/2061096


"""

import pandas as pd
import numpy as np


def entropy(probabilities):
    """
    Calculate the entropy of a list of probabilities.
    """
    p = [p for p in probabilities if p > 0]
    return -np.sum(p * np.log(p))

def mutual_information(df, group="g", unit="u", n="n"):
    data = df.copy()

    N = data[n].sum()
    # marginal probabilities
    data['p_ug'] = data[n] / N
    p_u = data.groupby(unit)[n].sum() / N  # each unit
    p_g = data.groupby(group)[n].sum() / N  # each group

    data['p_u'] = data[unit].map(p_u)
    data['p_g'] = data[group].map(p_g)
    M = np.sum(data['p_ug'] * np.log(data['p_ug'] / (data['p_u'] * data['p_g'])))
    E = entropy(p_g)
    H = M / E
    return M, H

def dissimilarity(df, cat, group="g", unit="u", n="n"):
    """Calculate the dissimilarity index for a group
    compared to the non-group population.

    Parameters
    ----------
    df : DataFrame
        A DataFrame in "long" format with columns for the group (g), the unit (u), and the
        total population (n). Each row represents the population at a group-unit intersection.
        In our school data, the unit would be the school (identified by DBN) the group
        would be one of the racial/ethnic categories (black, asian) or other demographic
        category (poverty, ELL, SWD) and the population would be the enrollment number
        of that group at that school (40 ELL students at 13K009).
    cat : str
        The category to compare to the non-group population. For example, "white" or "black".
        If there are four distinct values in the group column (e.g. "white", "black", "hispanic", "asian"),
        and cat is "white", then the dissimilarity index will be calculated for the white population
        against the total non-white population.
    group : str
        The name of the column in `df` that contains the group labels.
    unit : str
        The name of the column in `df` that contains the unit labels. (e.g. "school", "zipcode", "tract", "state")
    n : str
        The name of the column in `df` that contains the population counts.

    Returns
    -------
    float
        The dissimilarity index (D) for the group compared to the non-group population.
        D is a number between 0 and 1, where 0 indicates perfect integration and 1 indicates
        complete segregation.
    
    """
    data = df.copy()

    N = data[n].sum() # total population
    group_data = data[data[group] == cat] # just the rows for our group
    G = group_data[n].sum() # group population
    X = N - G # non-group population

    G_i = group_data[n]  # group at each location
    N_u = data.groupby(unit)[n].sum()
    N_u = group_data[unit].map(N_u)
    X_i = N_u - G_i

    P_g = G_i / G
    P_o = X_i / X
    D = np.sum(np.abs(P_g - P_o)) / 2

    return D




from nycschools import config, dataloader as loader
import requests
import os
import pytest
import shutil


tmp_data_dir = "/tmp/nyc-schools-test-data"

@pytest.mark.skip(reason="too slow for normal testing")
def test_read_urls():
    """Reads the urls download data file and makes sure they can all be reached."""
    assert "demographics" in config.urls
    for key, link in config.urls.items():
        print(f"checking {key}")
        r = requests.get(link.url)
        assert r.status_code == 200


@pytest.mark.skip(reason="deprecated")
def test_download_source():
    loader.download_source()

def test_contains_data_files():

    assert loader.contains_data_files(config.data_dir), "data files should be present in the default data directory"

    if os.path.exists(tmp_data_dir):
        shutil.rmtree(tmp_data_dir)
    os.mkdir(tmp_data_dir)
    assert not loader.contains_data_files(tmp_data_dir)
    shutil.rmtree(tmp_data_dir)


@pytest.mark.skip(reason="too slow for normal testing")
def test_download_archive():
    
    if os.path.exists(tmp_data_dir):
        shutil.rmtree(tmp_data_dir)

    os.mkdir(tmp_data_dir)
    loader.download_archive(data_dir=tmp_data_dir)
    assert loader.contains_data_files(tmp_data_dir)
    shutil.rmtree(tmp_data_dir)





import pytest
import pandas as pd
import os.path
from nycschools import config, nysed
from test_exams import check_state_test

@pytest.mark.skip(reason="too slow for normal testing")
def test_load_nysed_ela_math_archives():
    df = nysed.load_nysed_ela_math_archives()
    check_state_test(df)

    tmp = os.path.join(config.data_dir, "tmp")
    assert not os.path.exists(tmp), "temp files not cleaned up"

    feather = os.path.join(config.data_dir, "nysed-exams.feather")
    df = pd.read_feather(feather)
    check_state_test(df)

    csv = os.path.join(config.data_dir, "nysed-exams.csv")
    df = pd.read_csv(csv)
    check_state_test(df)


def test_load_nys_nysed():
    df = nysed.load_nys_nysed()
    n = len(df)
    assert n > 2_400_000, f"Only ({n}) records found"
    check_state_test(df)


def test_load_nyc_nysed():
    df = nysed.load_nyc_nysed()
    n = len(df)
    assert n > 680_000, f"Only ({n}) records found"
    check_state_test(df)



# NYC School Data
# Copyright (C) 2022-23. Matthew X. Curinga
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU AFFERO GENERAL PUBLIC LICENSE (the "License") as
# published by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the License for more details.
#
# You should have received a copy of the License along with this program.
# If not, see <http://www.gnu.org/licenses/>.
# ==============================================================================

import os.path

import pandas as pd

from .dataloader import load
from . import config

def load_admission_offers():
    return load(config.urls["shsat_apps"].filename)


def save_administration_offers():
    data_urls = config.urls["shsat_apps"].data_urls

    def shsat(ay, url):
        data = pd.read_csv(url)
        data.columns = ["dbn", "feeder_name", "hs_applicants_n", "testers_n", "offers_n"]
        data["ay"] = ay
        data["hs_applicants_n"] = data.hs_applicants_n.apply(lambda x: 2 if x == "0-5" else x)
        data["offers_n"] = data.offers_n.apply(lambda x: 2 if x == "0-5" else x)
        data["testers_n"] = data.testers_n.apply(lambda x: 2 if x == "0-5" else x)
        return data

    df = pd.concat([shsat(url, ay) for url, ay in data_urls.items()])
    df.hs_applicants_n = pd.to_numeric(df.hs_applicants_n, errors='coerce').astype('Int64')
    df.testers_n = pd.to_numeric(df.testers_n, errors='coerce').astype('Int64')
    df.offers_n = pd.to_numeric(df.offers_n, errors='coerce').astype('Int64')
    df["offers_pct"] = df.offers_n / df.testers_n
    f = os.path.join(config.data_dir, config.urls["shsat_apps"].filename)
    df.to_csv(f, index=False)
    return df


# NYC School Data
# Copyright (C) 2022. Matthew X. Curinga
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU AFFERO GENERAL PUBLIC LICENSE (the "License") as
# published by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the License for more details.
#
# You should have received a copy of the License along with this program.
# If not, see <http://www.gnu.org/licenses/>.
# ==============================================================================
import pandas as pd
from IPython.display import Markdown as md
from IPython.display import display
from decimal import *

import matplotlib.pyplot as plt
import matplotlib as mpl
import seaborn as sns
import networkx as nx
import folium
import numpy as np
import random
import math
from shapely import Point
from functools import partial


def ul(t):

    items =  [f"- {i}" for i in t]
    return str("\n".join(items))


def label_shapes(m, df, col, style={}):
    """Create a function that will add the string of `col`
    to the center of each shape in df """
    style_str = ";".join([f"{k}:{v}" for k,v in style.items()])
    def label(row):  
        point = row.geometry.centroid
        html=f"""<div style="{style_str}">{row[col]}</div>"""
        folium.Marker(
            location=(point.y, point.x), 
            icon=folium.DivIcon(html=html)).add_to(m)
    df.apply(label, axis=1)
    return m

def map_legend(m, items, title="", position="topleft"):
    """Create an interactive legend for the map with the items listed
    in the order they are passed in. The items should be a list
    of tuples with the first element being the label, the second
    being the html color, and (optional) third being the layer name.
    If a layer is included, the when that legend item is clicked, 
    items in that layer will be highlighted (toggled).


    Parameters
    ----------
    m: folium.Map the map to add the legend to
    items: list of tuples
        The items to include in the legend in the format (label, color)
        they will be displayed in the order they are passed in.
        For example: [("label1", "red"), ("label2", "blue"), ("label3", "#00ff00")]
        or [("label1", "red", "layer1"), ("label2", "blue", "layer2"), ("label3", "#00ff00", "layer3")]
        If the first tuple has 3 items, it is assumed that the legend will be
        interactive to toggle layers, i.e. each item **must** have a layer.
    title: str
        The title of the legend, appears at the top of the legend (default is "").
    position: string
        topleft | topright | bottomleft | bottomright (default is "topleft")
        Where to position the legend on the map.
    
    Returns
    -------
    folium.Map
        The map with the legend added to it.
    """

    css = """
.MapLegend {
  position: absolute;
  top: 10px;
  left: 80px;
  width: 200px;
  max-height: 600px;
  background: rgba(255,255,255,.8);
  z-index: 1000;
  padding: 1em;
  border: 2px solid lightgray;
  border-radius: 8px;
  overflow: auto;
  font-size: 14px;
  font-weight: bold;
}
.LegendItem {
    display: flex; 
    align-items: start; 
    margin-bottom: .5em;
}
.LegendMarker {
    width: 12px;
    height: 12px;
}

.LegendLabel {
    line-height: 14px; 
    padding-left: .25em; 
    font-size: 12px;
}
.MapLegend .clickable {
    cursor: pointer;
}
    """
    
    def legend_item(label, color, layer=None):
        onclick = ""
        clickable = ""
        if layer:
            onclick = f"toggleLayer('{layer}')"
            clickable = " clickable"
        return f"""
        <div class="LegendItem">
          <div class="LegendMarker{clickable}" {onclick} style="background:{color};">&nbsp;</div>
          <div class="LegendLabel{clickable}" {onclick}><strong>{label}</strong></div>
        </div>
"""
    
    if len(items[0]) == 2:
        legend_items = "".join([legend_item(label, color) for label, color in items])
    elif len(items[0]) == 3:
        legend_items = "".join([legend_item(label, color, layer) for label, color, layer in items])
    else:
        raise ValueError("Legend items must be a list of tuples with 2 or 3 elements.")

    html = f"""
<div class="MapLegend">
  <style>{css}</style>
  <h4><strong>{title}</strong></h4>
  {legend_items}
</div>
"""

    m.get_root().html.add_child(folium.Element(html))
    return m

def map_layers(m, df, radius=5):

    def create_layer(df, name, color="color", popup="popup", radius=5):
        layer = folium.FeatureGroup(name=name)
        def marker(row):
            if popup in row:
                info = row[popup]
            else:
                info = ""
            return folium.Circle(
                location=(row['geometry'].y, row['geometry'].x),
                radius=20,
                color=row["color"],
                fill=True,
                fill_color=row[color],
                fill_opacity=1,
                opacity=1,
                popup=info,
                className=f"layer-marker layer-{name} zoomable"
            )
        df.apply(lambda row: marker(row).add_to(layer), axis=1)

        layer.add_to(m)

        return layer

    groups = df.groupby("layer")
    for name, group in groups:
        create_layer(group, name, radius=radius)
    # add layer control
    # folium.LayerControl().add_to(m)
    return m


def rand_points(geometry, n, max_conflicts=100):
    """
    Create n random points within the bounds of the geometry.
    
    Parameters
    ----------
    geometry: shapely.geometry
        The geometry to create points within
    n: int
        The number of points to create

    Returns
    -------
    GeoDataFrame
        A GeoDataFrame with n random points within the geometry

    """
    minx, miny, maxx, maxy = geometry.bounds

    occupied = set()
    conflicts = 0

    def is_occupied(p):
        if p in occupied:
            conflicts += 1
            return True or conflicts > max_conflicts
        occupied.add(p)
        return False

    def rand_point():

        x = random.uniform(minx, maxx)
        y = random.uniform(miny, maxy)
        p = Point(x, y)
        if p.within(geometry):
            return p
        return rand_point()

    points = [rand_point() for i in range(n)]
    if conflicts > 0:
        print(f"Conflicts in plotting {n} points:", conflicts)
    # create a GeoDataFrame from the points
    points = pd.DataFrame({"geometry": points})
    return points

def map_js(m, file_path, js):
    """Add the custom javascript to the map
    that will load after the body is rendered
    and all other map elements are ready.
    Do not wrap js in <script> tags."""
    m.save(file_path)

    html = """
    <script>
    document.addEventListener("DOMContentLoaded", function() {""" + js + """

    });
    </script>
    """

    # open the map html and insert the footer at the bottom of page
    with open(file_path, 'r') as file:
        content = file.read()

    with open(file_path, 'w') as file:
        file.write(content.replace('</html>', f"{html}</html>"))
    
    return file_path


def map_footer(m, file_path, html):
    """Add the html to the bottom of the map html file"""
    m.save(file_path)

    # open the map html and insert the footer at the bottom of page
    with open(file_path, 'r') as file:
        content = file.read()

    with open(file_path, 'w') as file:
        file.write(content.replace('</body>', f"{html}</body>"))
    
    return file_path

def map_header(m, file_path, html):
    """Add the html to the top of the map html file"""
    m.save(file_path)

    # open the map html and insert the header at the top of page
    with open(file_path, 'r') as file:
        content = file.read()

    with open(file_path, 'w') as file:
        file.write(content.replace('<body>', f"<body>{html}"))
    
    return file_path


def create_layer(m, data, name, style={"color": "blue", "weight": 0, "opacity": 0},tooltip=None,radius=2):
    """Create a folium layer from the data. A new CircleMarker is created for
    each row in the DataFrame and added to map `m`.
    Parameters
    ----------
    m: folium.Map
        The map to add the layer to
    data: GeoDataFrame
        The data to plot
    name: str
        The name of the layer
    style: dict
        The style of the layer. Default is {"color": "blue", "weight": 0, "opacity": 0}
    tooltip: str
        The column name to use for the tooltip. Default is None.
    radius: int
        The radius of the circle markers. Default is 2.

    Returns
    -------
    folium.FeatureGroup
        The layer created from the data
    """
    layer = folium.FeatureGroup(name=name)

    def marker(row):
        tool = row[tooltip] if tooltip else False
        return folium.CircleMarker(
            location=(row['geometry'].y, row['geometry'].x),
            radius=radius,
            style=style,
            fill=True,
            popup=tool
        )
    data.apply(lambda row: marker(row).add_to(layer), axis=1)

    layer.add_to(m)

    return layer


def popup(cols, style={"min-width": "200px"}, fmt_funcs={}):
    style_str = ";".join([f"{k}:{v}" for k,v in style.items()])

    def html(row):
        def content(c):
            if c == "----":
                return """<hr style="padding: 0;margin:0; margin-bottom: .25em; border: none; border-top: 2px solid black;">"""
            # make a partial of fmt_num that includes the column name

            f = fmt_funcs.get(c, partial(fmt_num, c))
            return f"{nice_name(c)}: {f(row[c])}<br>"
        
        items = [content(c) for c in cols]
        items[0] = f"<strong>{items[0]}</strong>"
        items = "".join(items)
        return f'<div style="{style_str}">{items}</div>'

    return html


def fmt_num(col, n):
    if col.endswith("_pct"):
        return pct(n)
    try:
        n = float(n)
        if round(n) == n:
            return f"{int(n):,}"
        else:
            return f"{n:,.2f}"
    except:
        return n

def hexmap(cmap):

    def f(color):
        return mpl.colors.rgb2hex(cmap(color))

    return f

def pct(n):
    try:
        whole = int(n)
        if whole == float(n):
            return f"{whole}%"
        n = float(n)
        return f"{n:.1%}"
    except:
        return "-"

def commas(n):
    try:
        float(n)
        return f"{round(n, 3):,}"
    except:
        return "-"


def fmt_table(df, col_map=None, pct_cols=[], num_cols=[]):
    result = df.copy()
    for col in pct_cols:
        result[col] = result[col].apply(pct)

    for col in num_cols:
        result[col] = result[col].apply(commas)

    if col_map:
        result = result.rename(columns=col_map)
    return result


def show_md(s):
    display(md(s))

def infinite():
    n = 0
    while True:
        n += 1
        yield n


def counter():
    x = infinite()
    return x.__next__

# strips the leading zero from a rounded float
def round_f(f, places):

    s = str(round(f, places))
    if not "." in s:
        return f

    whole, frac = s.split(".")
    if whole == "0":
        whole = ""
    frac = frac.ljust(places, "0")
    return f"{whole}.{frac}"


def fmt_pearson(r):
    """Formats the Pearson's R correlation table returned
    from `pengouin.corr` in the format r(df)={r}, p={p}.
    The r is rounded to 2 decimals, and p is rounded to 3 decimals.
    """
    df = r.n[0] - 2
    p = round_f(r['p-val'][0], 3)
    r_val = round_f(r['r'][0], 2)
    return f"r({df})={r_val}, p={p}"


def edge_label(p, r):
    return f"{p}={round_f(r,2)}"

def nice_name(n):
    allcaps = ["dbn", "beds"]
    if n in allcaps:
        return n.upper()

    return n.replace("_", " ").title()

def plot_model(model):


    # get the data we need from model
    dv = model.model.endog_names

    params = list(model.params.index.values[1:])
    coefs = list(model.params.values[1:],)
    pvalues = list(model.pvalues.round(3).values[1:])


    network_map(dv, params, coefs, pvalues=None)

def network_map(dv, params, coefs, pvalues):

    if not pvalues:
        pvalues = [0 for i in coefs]

    cmap = mpl.cm.seismic
    nodes =  params + [dv]
    plabels = [f"p={p}" for p in pvalues]
    pnodes = list(zip(plabels, params))

    # all targets point to the dv
    targets =  [dv for _ in range(len(params))]
    edges = list(zip(params, targets))
    colors = coefs + [0]

    weights =  [abs(c) for c in coefs]
    weighted_edges = list(zip(params, targets, weights))

    max_node = 2000
    max_size = max(weights)
    node_size = [math.ceil(max_node * (c/max_size)) for c in weights]
    node_size.append(  max_node * 2 )

    labels =  dict([(n, nice_name(n)) for n in nodes])

    edge_labels = dict([(x, edge_label(x[0], y)) for x, y in zip(edges, coefs)])
    node_dict = {}
    for i, n in enumerate(nodes):
        node_dict[n] = {
            "node_size": node_size[i],
            "color": colors[i],
            "pvalue": pvalues[i] if i < len(pvalues) - 1 else 0,
            "weight": weights[i] if i < len(weights) - 1 else 0,
            "scale": node_size[i] / node_size[i],
            "label": nice_name(n)
        }

    draw_model(nodes, pnodes, node_size, weighted_edges, labels, edge_labels, colors, cmap, node_dict)

def draw_model(nodes, pnodes, node_size, edges, labels, edge_labels, colors, cmap, node_dict):

    G = nx.DiGraph()
    G.add_nodes_from(node_dict)
    G.add_weighted_edges_from(edges)

    fig, ax = plt.subplots(figsize=(16,9))

    # pos = nx.spring_layout(G, k=3)
    pos = nx.circular_layout(G)

    nx.draw(G, pos=pos, ax = ax, with_labels=False, node_color=colors, cmap=cmap, node_size=node_size,
            linewidths=2, min_source_margin=2, min_target_margin=2, font_size=14)


    nx.draw_networkx_edge_labels(G,pos, ax=ax, edge_labels=edge_labels, label_pos=.5, font_size=12)

    # draw the node labels separately to put below the nodes
    pos_attrs = {}
    for node, coords in pos.items():

        x, y = coords
        pos_attrs[node] = (x, y - .1)

    nx.draw_networkx_labels(G, pos_attrs, labels=labels)


    # adding p-values as node
    # G.add_nodes_from(pnodes,{"pvalue":True})
    # nx.draw_networkx_nodes(G, pos, ax=ax, nodelist=G.nodes(data=True))




    plt.tight_layout()
    plt.margins(0.05)
    plt.show()



# NYC School Data
# Copyright (C) 2022. Matthew X. Curinga
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU AFFERO GENERAL PUBLIC LICENSE (the "License") as
# published by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the License for more details.
#
# You should have received a copy of the License along with this program.
# If not, see <http://www.gnu.org/licenses/>.
# ==============================================================================
import pandas as pd
import numpy as np
import wget

import os.path
import zipfile

from . import schools
from . import config
import shutil

urls = config.urls


def load_snapshots():
    """
    Loads the "snapshot" data from the NYC DOE data portal.
    This data contains core columns only for all "school types"
    mainly focusing on school demographics and teacher demographics.
    """
    path = os.path.join(config.data_dir, "nysed-exams.feather")
    return pd.read_feather(path)



from nycschools import class_size
import ssl

# Temporarily disable SSL certificate verification
ssl._create_default_https_context = ssl._create_unverified_context

def test_get_class_size():
    df = class_size.get_class_size()
    cols = {'ay', 'dbn', 'grade', 'students_n', 'classes_n', 'avg_class_size',
            'min_class_size', 'max_class_size', 'dept', 'subject'}
    assert cols.issubset(df.columns)
    assert df.ay.nunique() >= 2


def test_load_class_size():
    df = class_size.load_class_size()

    cols = {'dbn', 'school_name', 'grade', 'program_type', 'students_n',
       'classes_n', 'avg_class_size', 'min_class_size', 'max_class_size',
       'dept', 'subject', 'ay'}
    assert cols.issubset(df.columns)
    assert df.ay.nunique() >= 2

def test_load_ptr():
    df = class_size.load_ptr()
    cols = {'dbn', 'school_name', 'ptr', 'ay'}
    assert cols.issubset(df.columns)




# NYC School Data
# Copyright (C) 2022. Matthew X. Curinga
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU AFFERO GENERAL PUBLIC LICENSE (the "License") as
# published by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the License for more details.
#
# You should have received a copy of the License along with this program.
# If not, see <http://www.gnu.org/licenses/>.
# ==============================================================================
import pandas as pd
import numpy as np

import os.path
from concurrent.futures import ThreadPoolExecutor

from .dataloader import load
from . import config
urls = config.urls


# convert these to numbers or coerce to NaN
numeric_cols = [
    'number_tested',
    "mean_scale_score",
    "level_1_n",
    "level_1_pct",
    "level_2_n",
    "level_2_pct",
    "level_3_n",
    "level_3_pct",
    "level_4_n",
    "level_4_pct",
    "level_3_4_n",
    "level_3_4_pct"
]


def load_charter_ela(url=urls["charter_ela"].url):
    """
    Loads the charter school ELA exam results for the "All Students"
    category from the NYC Open Data Portal. Columns are re-named for consistency.
    """
    filename = urls["charter_ela"].filename
    return load_charter_test(url, filename)

def load_charter_math(url=urls["charter_math"].url):
    """
    Loads the charter school Math exam results for the "All Students"
    category from the NYC Open Data Portal. Columns are re-named for consistency.
    """
    filename = urls["charter_math"].filename
    return load_charter_test(url, filename)

def load_charter_test(url, filename):
    try:
        return load(filename)
    except FileNotFoundError:
        pass

    df = pd.read_csv(url)
    df = charter_cols(df)
    out = os.path.join(config.data_dir, filename)
    df.to_csv(out, index=False)

    return df

def load_math_ela_wide():
    """
    Load a combined `DataFrame` with both math and ela test results
    in a "wide" data format.
    All of the math result columns have the suffix `_math`
    and the ELA columns have the suffix `_ela`.
    """
    math_df = load_math()
    ela_df = load_ela()
    combined = math_df.merge(ela_df, how="inner", on=["dbn", "ay", "grade"], suffixes=["_math", "_ela"])
    return combined

def load_math_ela_long():
    """
    Load a combined `DataFrame` with both math and ela test results
    in a "wide" data format.
    All of the math result columns have the suffix `_math`
    and the ELA columns have the suffix `_ela`.
    """
    math_df = load_math().copy()
    ela_df = load_ela().copy()
    math_df["exam"] = "math"
    ela_df["exam"] = "ela"
    return pd.concat([math_df, ela_df])


def load_ela():
    """
    Loads the New York State ELA grades 3-8 ELA exam results for all categories.
    If a local .csv data file exists, it will return results from that file. If
    no local file is available, itt will loads the Excel data file for from the
    NYC Data Portal and then cobmines the results with charter school data into
    a `DataFrame`. _This can be slow_.

    """
    # filename = os.path.join(config.data_dir, urls["nyc_ela"].filename)
    # df = pd.read_csv(filename, low_memory=False)
    df = load(urls["nyc_ela"].filename)
    return df.sort_values(by=["dbn", "ay"])


def load_math():
    """
    Loads the New York State Math grades 3-8 ELA exam results for all categories.
    If a local .csv data file exists, it will return results from that file. If
    no local file is available, itt will loads the Excel data file for from the
    NYC Data Portal and then combines the results with charter school data into
    a `DataFrame`. _This can be slow_.
    """
    # filename = os.path.join(config.data_dir, urls["nyc_math"].filename)
    # df = pd.read_csv(filename, low_memory=False)
    df = load(urls["nyc_math"].filename)

    return df.sort_values(by=["dbn", "ay"])


def load_regents():
    """
    Loads the New York State Regents exam scores for all categories.
    @return `DataFrame`
    """
    return load(urls["nyc_regents"].filename)


# ==============================================================================

def charter_cols(data):
    df = data.copy()

    if "unnamed_column" in df:
        del df["unnamed_column"]

    df["ay"] = df["year"] - 1

    new_cols = {
        'year': 'test_year',
        'level_1': 'level_1_n',
        'level_1_1': 'level_1_pct',
        'level_2': 'level_2_n',
        'level_2_1': 'level_2_pct',
        'level_3': 'level_3_n',
        'level_3_1': 'level_3_pct',
        'level_4': 'level_4_n',
        'level_4_1': 'level_4_pct',
        'level_3_4': 'level_3_4_n',
        'level_3_4_1': 'level_3_4_pct'
    }

    df = df.rename(columns=new_cols)

    df["charter"] = 1
    for col in numeric_cols:
        df[col] = pd.to_numeric(df[col], errors='coerce')
        if col.endswith("pct"):
            df[col] = df[col] / 100

    return df


def load_regents_excel():
    url = urls["nyc_regents"].url
    xls = pd.read_excel(url, sheet_name=None)
    cols = [
        'School DBN',
        'School Type',
        'School Level',
        'Regents Exam',
        'Year',
        'Category',
        'Total Tested',
        'Mean Score',
        'Number Scoring Below 65',
        'Percent Scoring Below 65',
        'Number Scoring 65 or Above',
        'Percent Scoring 65 or Above',
        'Number Scoring 80 or Above',
        'Percent Scoring 80 or Above',
        'Number Scoring CR',
        'Percent Scoring CR']

    new_cols = [
        'dbn',
        'school_type',
        'school_level',
        'regents_exam',
        'year',
        'category',
        'number_tested',
        'mean_score',
        'below_65_n',
        'below_65_pct',
        'above_64_n',
        'above_64_pct',
        'above_79_n',
        'above_79_pct',
        'college_ready_n',
        'college_ready_pct']

    sheet_names = ['All Students',
        'By Gender',
        'By Ethnicity',
        'By ELL Status',
        'By SWD Status']

    data = [xls[sheet] for sheet in sheet_names]
    df = pd.concat(data, ignore_index=True)
    df = df[cols]
    df.columns = new_cols
    numeric_cols = new_cols[6:]
    for col in numeric_cols:
        df[col] = pd.to_numeric(df[col], errors='coerce')
    df["test_year"] = df["year"]
    df["ay"] = df["year"] - 1
    df = df.sort_values(by=["dbn","ay","regents_exam","category"])
    filename = os.path.join(config.data_dir, urls["nyc_regents"].filename)
    df.to_csv(filename, index=False)
    return df




def read_nys_exam_excel(url):
    """Reads the NYS exam results from an Excel file.
    The Excel file contains multiple sheets, each of which
    contains a different demographic breakdown of the exam.
    The excel files contain more recent data than the data portal
    or NYSED downloads, however they do not contain demographic
    breakdowns for charter schools.

    Note this function can take more than 120 seconds to run.
    
    Parameters:
        url (str): the URL of the Excel file
    Returns:
        `DataFrame`
    """

    xls = pd.read_excel(url, sheet_name=None)
    
    # these are the known sheet names that we care about
    sheet_names = ['All', 'SWD', 'Ethnicity', 'Gender', 'Econ Status', 'ELL']

    data = [xls[sheet] for sheet in sheet_names]

    df = pd.concat(data, ignore_index=True)

    cols = ['DBN', 'Grade', 'Year', 'Category', 'Number Tested', 'Mean Scale Score',
            '# Level 1', '% Level 1', '# Level 2', '% Level 2', '# Level 3',
            '% Level 3', '# Level 4', '% Level 4', '# Level 3+4', '% Level 3+4']

    new_cols = ['dbn', 'grade', 'year', 'category', 'number_tested',
                'mean_scale_score', 'level_1_n', 'level_1_pct', 'level_2_n', 'level_2_pct',
                'level_3_n', 'level_3_pct', 'level_4_n', 'level_4_pct', 'level_3_4_n',
                'level_3_4_pct']

    df = df[cols]

    df.columns = new_cols

    for col in numeric_cols:
        df[col] = pd.to_numeric(df[col], errors='coerce')
        if col.endswith("pct"):
            df[col] = df[col] / 100

    df["test_year"] = df["year"]
    df["ay"] = df["year"] - 1
    del df["year"]
    df["charter"] = df["dbn"].str.startswith("84")
    return df


def load_math_excel(url=urls["nyc_math"].url):
    """
    Load NYS Math test scores from the Excel workbook rather than the API. NYC DOE
    fails to release the demographic breakdowns for test results via its open
    data api.
    """

    df = read_nys_exam_excel(url)
    charter_df = load_charter_math()
    df = pd.concat([df, charter_df])
    filename = os.path.join(config.data_dir, urls["nyc_math"].filename)
    df.to_csv(filename, index=False)
    return df

def load_ela_excel(url=urls["nyc_ela"].url):
    """
    Load NYS ELA test scores from the Excel workbook rather than the API. NYC DOE
    fails to release the demographic breakdowns for test results via its open
    data api.
    """

    df = read_nys_exam_excel(url)
    charter_df = load_charter_ela()
    df = pd.concat([df, charter_df])
    filename = os.path.join(config.data_dir, urls["nyc_ela"].filename)
    df.to_csv(filename, index=False)
    return df



# NYC School Data
# Copyright (C) 2022. Matthew X. Curinga
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU AFFERO GENERAL PUBLIC LICENSE (the "License") as
# published by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the License for more details.
#
# You should have received a copy of the License along with this program.
# If not, see <http://www.gnu.org/licenses/>.
# ==============================================================================
from nycschools import datasets
import pandas as pd
import geopandas as gpd
from bs4 import BeautifulSoup
import requests
import os
import os.path
from datetime import datetime


from . import config
from nycschools.dataloader import load

urls = config.urls
school_location_file = os.path.join(config.data_dir, urls["school_locations"].filename)

def load_zipcodes():
    """Load the NYC zip code boundaries as a GeoDataFrame from data_dir.
    Zip codes are compiled from the NYC Data Portal via the US Post Office"""
    df = load(urls["zipcodes"].filename)
    return df

def load_school_locations():
    """Returns a GeoDataFrame with the school locations and location meta-data"""

    try:
        df = load(urls["school_locations"].filename)
        return df
    except Exception as e: # geopandas throws DriveError, but I don't know where to import it to catch it
        if e.type != "<class 'fiona.errors.DriverError'>":
            raise e

def load_school_geo_points():
    """Load only the school location points as a GeoDataFrame"""
    df = load_school_locations()
    return df[["dbn", "x", "y", "geometry"]]

def get_and_save_locations(filename=school_location_file):
    points = get_points()
    locations = get_locations()
    df = points.merge(locations, on="dbn", how="left")
    # df = gpd.GeoDataFrame(points)
    df.open_year = df.open_year.fillna(0)
    df.open_year = df.open_year.astype(int)
    out = df.copy()
    out.to_file(filename, driver="GeoJSON")
    return df

def get_points(geojsonurl=urls["school_geo"].url):
    """Read the school location points and zipcodes from an Open Data Portal GeoJSON URL"""
    # this is the API feed for the location points
    # it's the best place to get zip codes

    df = gpd.read_file(geojsonurl)
    df = df.drop_duplicates()

    df = df.rename(columns={"xcoordinat":"x","ycoordinat":"y",})
    df.x = pd.to_numeric(df.x, errors='coerce')
    df.y = pd.to_numeric(df.y, errors='coerce')
    df = df[df.x > 0]
    df.bbl = df.bbl.astype(str)
    df.bbl = df.bbl.str.replace(".0", "")
    df["dbn"] = df.ats_code
    df["district"] = df.adimindist.astype(int)
    df = df.rename(columns={"geodistric":"geo_district"})
    cols = ['dbn', 'zip', 'geo_district', 'district', 'x','y', 'bbl', 'geometry']

    df = df[cols]
    df.zip = df.zip.astype("string")
    return df

def get_locations(url=urls["school_locations"].url):
    """
       Read school level data with many location-related columns: school x,y
       coords, and data about the school locations including NYS BEDS ids,
       census tract, and police precinct.
    """

    locations = pd.read_csv(url)
    locations = locations.drop_duplicates()
    locations["dbn"] = locations.system_code
    cols = [
        'dbn',
        'administrative_district_code',
        'administrative_district_name',
        'beds',
        'borough_block_lot',
        'census_tract',
        'community_district',
        'community_school_sup_name',
        'council_district',
        'fax_number',
        'fiscal_year',
        'geographical_district_code',
        'grades_final_text',
        'grades_text',
        'highschool_network',
        'highschool_network_location',
        'highschool_network_name',
        'latitude',
        'location_category_description',
        'location_code',
        'location_name',
        'location_type_description',
        'longitude',
        'managed_by_name',
        'nta',
        'nta_name',
        'open_date',
        'police_precinct',
        'primary_building_code',
        'principal_name',
        'principal_phone_number',
        'principal_title',
        'state_code',
        'status_descriptions']

    locations = locations[cols]
    locations.beds = locations.beds.astype("string")
    locations.open_date = locations.open_date.fillna(0)
    locations.open_date = locations.open_date.apply(lambda x: int(str(x).split('-')[0] if x != 0 else 0))
    locations.rename(columns={"open_date": "open_year"}, inplace=True)
    return locations

def load_school_footprints():
    gdf = load(urls["building_footprints"].school_footprints_file)
    return gdf

def load_city_footprints():
    path = os.path.join(config.data_dir, urls["building_footprints"].city_footprints_feather)
    gdf = gpd.read_file(path)

    return gdf


def get_school_footprints():
    """Load all building footprints from the NYC Dept of Buildings
    Open Data Portal. Merge these shapes with school locations
    and save the footprints"""
    print("Getting footprints from URL")
    url=urls["building_footprints"].url
    foot = gpd.read_file(url)
    foot["bbl"] = foot.base_bbl.astype(str)
    foot.mpluto_bbl = foot.mpluto_bbl.astype(str)
    # saving local feather
    city_path = os.path.join(config.data_dir, urls["building_footprints"].city_footprints_feather)
    foot.to_feather(city_path)

    # merging with point locations for dbn
    school_loc = get_points()
    school_foot = foot.merge(school_loc[["dbn","bbl"]], on="bbl", how="inner")
    # saving to data dir
    school_path = os.path.join(config.data_dir, urls["building_footprints"].school_footprints_file)
    school_foot.to_file(school_path, driver="GeoJSON")
    return school_foot

def load_districts(url=urls["district_geo"].url):
    """Get geo shape file for NYC school districts, indexed by district number."""
    districts = gpd.read_file(url)
    # rename the columns
    districts.columns = ['district', 'area', 'length', 'geometry']
    districts.district = pd.to_numeric(districts.district, downcast='integer', errors='coerce')
    districts = districts.to_crs(epsg=4326)
    return districts




from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (here / "README.md").read_text(encoding="utf-8")

setup(
    packages=['nycschools'],

)



# NYC School Data
# Copyright (C) 2023. Matthew X. Curinga
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU AFFERO GENERAL PUBLIC LICENSE (the "License") as
# published by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the License for more details.
#
# You should have received a copy of the License along with this program.
# If not, see <http://www.gnu.org/licenses/>.
# ==============================================================================
import os.path
import re

import pandas as pd
import requests


from . import config, schools



def get_ceps():
    stem = config.urls["cep"].url_stem
    df = schools.load_school_demographics()
    df = df[df.ay == df.ay.max()]
    out_dir = os.path.join(config.data_dir, "ceps")
    os.makedirs(out_dir, exist_ok=True)
  
    def download(dbn):
        url = f"{stem}cep_{dbn[2:]}.pdf"
        out = os.path.join(out_dir, f"{dbn}.pdf")
        if not os.path.exists(out):
            r = requests.get(url)
            if r.status_code == 200:
                with open(out, "wb") as f:
                    f.write(r.content)
                return None
            else:
                print(f"Error downloading {dbn}: {r.status_code}")
                return dbn
    not_found = df.dbn.apply(download)


# NYC School Data
# Copyright (C) 2022. Matthew X. Curinga
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU AFFERO GENERAL PUBLIC LICENSE (the "License") as
# published by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the License for more details.
#
# You should have received a copy of the License along with this program.
# If not, see <http://www.gnu.org/licenses/>.
# ==============================================================================
import os
import os.path
import py7zr
import requests
import sys
import warnings
import platform
import subprocess
import time
import datetime
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor, as_completed

import pandas as pd
import geopandas as gpd

from yaspin import yaspin
from yaspin.spinners import Spinners

# from . import config, schools, exams, budgets, geo, nysed, shsat
from . import config
from .tools import suppress_warnings

import ssl

ssl._create_default_https_context = ssl._create_unverified_context

def get_data_dir():
    return config.data_dir


def read_file(path):
    print("reading", path)
    if path.endswith(".geojson"):
        return gpd.read_file(path)
    elif path.endswith(".csv"):
        return pd.read_csv(path)
    elif path.endswith(".feather"):
        return pd.read_feather(path)
    else:
        raise ValueError(f"Unknown file type: {path}")

def write_file(df, path):
    print("writing...", path)
    if path.endswith(".geojson"):
        df.to_file(path, driver="GeoJSON")
    elif path.endswith(".csv"):
        df.to_csv(path, index=False)
    elif path.endswith(".feather"):
        df.to_feather(path)
    else:
        raise ValueError(f"Unknown file type: {path}")

def load(path):
    print("loading...", path)
    remote_path = config.urls["datasite"].url + path
    if config.data_dir is None or config.data_dir == "":
        return read_file(remote_path)
    
    local_path = os.path.join(config.data_dir, path)
    if os.path.exists(local_path):
        return read_file(local_path)

    df = read_file(remote_path)
    write_file(df, local_path)
    return df


def download_file(url, local_filename):
    """Optimize file download using requests library."""
    with requests.get(url, stream=True) as r:
        r.raise_for_status()
        with open(local_filename, 'wb') as f:
            for chunk in r.iter_content(chunk_size=8192):
                f.write(chunk)
    return local_filename

def download_data():
    path = find_data_dir(config)
    if path:
        config.data_dir = path
        return path
    
    return download_archive(config.data_dir)


def find_data_dir(config):
    """
    Tries to find an existing data directory populated
    with data, including searching through mounted
    google drive if the `colab` package is available
    and the g drive is mounted in the "standard" location
    of `/content/gdrive`.

    Returns:
        str: the path to the data directory
    """
    env_dir = os.environ.get("NYC_SCHOOLS_DATA_DIR", None)
    local = os.path.join(".", "school-data")

    paths = [config.data_dir, env_dir, local]
    # first look locally
    for path in paths:
        if path and os.path.exists(path) and contains_data_files(path):
            config.data_dir = path
            return path
    # if not, try to mount a google drive
    path = mount_colab_data_dir()
    if path and contains_data_files(path):
        config.data_dir = path
        return path
    return None


def mount_colab_data_dir():
    """Try to mount a google drive directory in colab
    and then search for the data directory by looking
    for a directory with the known name 'nyc-schools-data'."""

    gdrive = "/content/gdrive"
    target = "nyc-schools-data"

    if not os.path.exists(gdrive):
        try:
            from google.colab import drive
            drive.mount(gdrive)
        except:
            return None

    # first check all of MyDrive
    for root, dirs, files in os.walk(f"{gdrive}/MyDrive"):
        if target in dirs:
            return os.path.join(root, f"{target}/data")

    # next check all of Shared Drives
    for root, dirs, files in os.walk(f"{gdrive}/Shareddrives"):
        if target in dirs:
            return os.path.join(root, f"{target}/data")

    return None



def contains_data_files(path):
    """
    Checks to see if the specified path contains
    the data files required by this application.

    Parameters:
        path (str): the path to check

    Returns:
        bool: True if the path contains the data files
            required by this application, False otherwise
    """
    files = set(os.listdir(path))
    if len(files) == 0:
        return False

    expected = {
        "charter-ela.csv",
        "charter-math.csv",
        "nyc-ela.csv",
        "nyc-math.csv",
        "nysed-exams.csv",
        "nysed-exams.feather",
        "school-demographics.csv",
        "school_locations.geojson"
    }

    if expected.issubset(files):
        return True
    
    missing = expected.difference(files)
    if len(missing) == len(expected):
        return False
    
    warnings.warn(f"""Some data files are missing from the data directory.
    Found files: {files.intersection(expected)}
    Missing files: {missing}
You can download the data files by running: `python -m nycschools.dataloader -d`
For more information, see:
https://adelphi-ed-tech.github.io/nycschools/""")

    
    return False



def download_archive(data_dir=None):
    """
    Downloads the school data archive to the local
    drive and saves it into `data_dir` then extracts
    the .7z archive. `data_dir` now contains the
    cleaned and compiled school data files.

    Parameters:
        data_dir (str): the path to the directory where
            the data files should be saved. If not specified,
            the package configuration `data_dir` is used.

    Returns:
        str: the path to the downloaded file
    """
    if not data_dir:
        print("no data dir")
        data_dir = config.data_dir
    else:
        config.data_dir = data_dir

    print("using data dir", data_dir)

    url = config.urls["school-data-archive"].url
    filename = config.urls["school-data-archive"].filename
    data_dir = os.path.abspath(data_dir)
    archive = os.path.join(data_dir, filename)
    resp = requests.get(url)
    with open(archive, "wb") as f:
        f.write(resp.content)

    with py7zr.SevenZipFile(archive, mode='r') as z:
        z.extractall(path=data_dir)

    os.remove(archive)
    return data_dir


def get_venv_activate():
    """Finds the activation script for a running virtual environment
    or `None` if not running a venv."""

    venv_path = os.environ.get('VIRTUAL_ENV')
    if venv_path:
        return os.path.join(venv_path, 'bin', 'activate')
    return None


def find_config_file():
    """Looks for virtual environment activation scripts
    or bash configuration files in known locations.
    
    Returns:
        str: the path to the configuration file or `None` if not found
    """
    paths = [
        get_venv_activate(),
        os.path.expanduser("~/.bashrc"),
        os.path.expanduser("~/.bash_profile"),
        os.path.expanduser("~/.profile")
    ] 
    
    for path in paths:
        if path and os.path.exists(path):
            return path
    
    warnings.warn(f"""Could not find a venv or bash configuration file to edit.
You should manually set the NYC_SCHOOLS_DATA_DIR environment variable.
See the full documentation at: {config.urls["docbook"].url}""")
    
    return None
    


def set_env_var(data_dir):
    """Attempts to set the NYC_SCHOOLS_DATA_DIR environment variable
    based on the user's platform."""

    # This will return 'Windows', 'Linux', or 'Darwin' (for macOS)
    os_type = platform.system()

    if os_type == "Windows":
        subprocess.run(["setx", "NYC_SCHOOLS_DATA_DIR", data_dir])
    elif os_type in ["Linux", "Darwin"]:
        config_path = find_config_file()
        if not config_path:
            return
        print("Writing env var to", config_path)
        with open(config_path, "a") as f:
            f.write(f"export NYC_SCHOOLS_DATA_DIR={data_dir}\n")
        print(f"""To access the data files in your current terminal session,
you must run the following command:
source {config_path}
""")



def download_cache():
    """Download the data archive and save it to the local drive.
    This interactive terminal program prompts the user for
    the location to save the data files. Once the files
    are downloaded and expanded it attempts to write the path
    to the data files into the python configuration environment.
    """
    print("Default data location:", config.data_dir)
    data_dir = ""

    def prompt_data():
        path = input("Enter the path to the data directory (or <enter> for default): ")
        if path == "":
            path = config.data_dir
        # check if it exists, if not create it, catch errors and re-prompt
        try:
            print("Creating data cache: ", path)
            os.makedirs(path, exist_ok=True)
        except:
            print(f"Could not create directory {path}")
            return prompt_data()
        
        # check if it's writeable
        if not os.access(path, os.W_OK):
            print(f"Cannot write to directory {path}")
            print("Either change file permissions or choose a different directory.")
            return prompt_data()
        return path

    data_dir = prompt_data()
    print(f"Downloading school data to {data_dir}")

    with yaspin(text="loading...", spinner=Spinners.bouncingBall) as sp:
        sp.side = "right"
        data_dir = download_archive(data_dir)
        sp.ok("✔")

    print(f"Data successfully saved to: {data_dir}")
    auto_config = input("Automatically set environment variable? [Y/n]: ")
    if auto_config.lower() == "y" or len(auto_config) == 0:
        set_env_var(data_dir)
        print("Environment variable set.")
    else:
        print("""You must configure the NYC_SCHOOLS_DATA_DIR environment variable.
See the full documentation at: {config.urls["docbook"].url}""")


def main():
    """Show the path to the `data_dir` where school data is stored.
To use the interactive downloader, run `python -m nycschools.dataloader -d`.
    """
    print(f"Current data directory: {config.data_dir}")

    # read args if exists
    if len(sys.argv) > 1:
        if sys.argv[1] == "-h" or sys.argv[1] == "--help":
            print("docs", main.__doc__)
            return
        elif sys.argv[1] == "-d" or sys.argv[1] == "--download":
            download_cache()
            return
        else:
            print("Unrecognized argument. Run `python -m nycschools.dataloader --help` for help.")
            return
    

if __name__ == "__main__":
    main()




# NYC School Data
# Copyright (C) 2022. Matthew X. Curinga
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU AFFERO GENERAL PUBLIC LICENSE (the "License") as
# published by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the License for more details.
#
# You should have received a copy of the License along with this program.
# If not, see <http://www.gnu.org/licenses/>.
# ==============================================================================
import os.path
import math
import numpy as np
import re

import pandas as pd
from thefuzz import fuzz

from . import geo
from . import config

from nycschools.dataloader import load


__demo_filename = os.path.join(config.data_dir, config.urls["demographics"].filename)

class demo():
    """The `demo` class bundles some common sets of column names
    to make it easier to work with the school demographic `DataFrame`

    Examples:
    ----------
    from nycschools import schools

    df = schools.load_school_demographics()
    basic = df[schools.demo.short_cols]
    basic.head()
    """
    raw_cols = [
       'dbn', 'school_name', 'year', 'total_enrollment',
       'grade_pk', 'grade_k', 'grade_1', 'grade_2', 'grade_3',
       'grade_4', 'grade_5', 'grade_6', 'grade_7', 'grade_8', 'grade_9',
       'grade_10', 'grade_11', 'grade_12', 'female', 'female_1', 'male',
       'male_1', 'asian', 'asian_1', 'black', 'black_1', 'hispanic',
       'hispanic_1', 'multi_racial', 'multi_racial_1', 'native_american',
       'native_american_1', 'white', 'white_1', 'missing_race_ethnicity_data',
       'missing_race_ethnicity_data_1', 'students_with_disabilities',
       'students_with_disabilities_1', 'english_language_learners',
       'english_language_learners_1', 'poverty', 'poverty_1',
       'economic_need_index'
    ]

    default_cols = [
        'dbn',
        'beds',
        'district',
        "geo_district",
        'boro',
        'school_name',
        'short_name',
        'ay',
        'year',
        'school_type',
        'total_enrollment',
        'grade_pk',
        'grade_k',
        'grade_1',
        'grade_2',
        'grade_3',
        'grade_4',
        'grade_5',
        'grade_6',
        'grade_7',
        'grade_8',
        'grade_9',
        'grade_10',
        'grade_11',
        'grade_12',
        'female_n',
        'female_pct',
        'male_n',
        'male_pct',
        'asian_n',
        'asian_pct',
        'black_n',
        'black_pct',
        'hispanic_n',
        'hispanic_pct',
        'multi_racial_n',
        'multi_racial_pct',
        'native_american_n',
        'native_american_pct',
        'white_n',
        'white_pct',
        'missing_race_ethnicity_data_n',
        'missing_race_ethnicity_data_pct',
        'swd_n',
        'swd_pct',
        'ell_n',
        'ell_pct',
        'poverty_n',
        'poverty_pct',
        'eni',
        'clean_name',
        'zip'
    ]

    core_cols = [
        'dbn',
        'beds',
        'district',
        "geo_district",
        'boro',
        'school_name',
        'ay',
        'total_enrollment',
        'female_n',
        'female_pct',
        'male_n',
        'male_pct',
        'asian_n',
        'asian_pct',
        'black_n',
        'black_pct',
        'hispanic_n',
        'hispanic_pct',
        'white_n',
        'white_pct',
        'swd_n',
        'swd_pct',
        'ell_n',
        'ell_pct',
        'poverty_n',
        'poverty_pct',
        'eni'
    ]

    short_cols = ["dbn","school_name", "short_name", "clean_name", "ay"]

    default_map = {
        'female':'female_n',
        'female_1':'female_pct',
        'male':'male_n',
        'male_1':'male_pct',
        'neither_female_nor_male_1': 'neither_female_nor_male_pct',
        'asian':'asian_n',
        'asian_1':'asian_pct',
        'black':'black_n',
        'black_1':'black_pct',
        'hispanic':'hispanic_n',
        'hispanic_1':'hispanic_pct',
        'multi_racial':'multi_racial_n',
        'multi_racial_1':'multi_racial_pct',
        'native_american':'native_american_n',
        'native_american_1':'native_american_pct',
        'white':'white_n',
        'white_1':'white_pct',
        'missing_race_ethnicity_data':'missing_race_ethnicity_data_n',
        'missing_race_ethnicity_data_1':'missing_race_ethnicity_data_pct',
        'students_with_disabilities':'swd_n',
        'students_with_disabilities_1':'swd_pct',
        'english_language_learners':'ell_n',
        'english_language_learners_1':'ell_pct',
        'poverty':'poverty_n',
        'poverty_1':'poverty_pct',
        'economic_need_index':'eni'
    }


def str_pct(row, pct_col, enroll_col):
    """generic function to take percentages represented as Strings and convert to real
    expects data to look like '84.33%' or .8433, 'Above 95%', 'Below 5%'
    Example: df["economic_need_index"] = df.apply(lambda row: str_pct(row, "economic_need_index", "total_enrollment"), axis = 1)
    """

    # just call the population size `n`
    n = row[enroll_col]

    # get the percentage as a string
    pct = str(row[pct_col])
    if pct.endswith("%"):
        pct = pct[:-1]
    if "Above" in pct:
        pct = n * .96 / n
    elif "Below" in pct:
        pct = n * .04 / n
    elif pct == "No Data":
        return 0

    try:
        pct = float(pct)
    except Exception as ex:
        print(
            f"Exception in {pct_col}: {row['dbn']} - {row['year']}. Value: {pct}")
        raise ex
    # some percentages are whole numbers 0..100, others are real numbers 0..1
    if pct >= 1:
        pct = pct / 100

    return float(pct)

def str_count(row, col, enroll_col):
    """generic function to take whole number counts represented as Strings and convert to integers
    expects data to look like '246', 'Above 95%', 'Below 5%'
    Example: df["poverty"] = df.apply(lambda row: str_count(row, "poverty", "total_enrollment"), axis = 1)
    """
    count = row[col]
    # just call the population size `n`
    n = row[enroll_col]
    try:
        return int(count)
    except:
        if "Above" in count:
            return int(math.ceil(n * .96))
        elif "Below" in count:
            return int(math.floor(n * .04))
        else:
            print(f"Exception in {col}: {row['dbn']} - {row['year']}. Value: {count}")

def school_type(school):
    """Any school that serves middle school kids is considered a middle school here."""

    if school["middle"]:
        return "MS"
    if school["elementary"]:
        return "PS"
    if school["hs"]:
        return "HS"
    return "NA"


def clean_name(sn):
    """Creates a simplified school name that is easier to search/index

    """
    sn = sn.lower()
    sn = sn.strip()
    sn = sn.replace(".", "")
    clean = []
    for word in sn.split(" "):
        try:
            n = int(word)
            clean.append(str(n))
        except:
            clean.append(word)

    sn = " ".join(clean)

    # ms, is, ps followed by a number
    p = re.compile(r"\b([m|p|i]s [0-9]*)")
    m = p.search(sn)
    if m:
        sn = sn.replace(m.group(0), "")
    sn = sn.strip()
    # if the only thing left is the boro, short name will be school num + boro
    if sn in ["bronx", "brooklyn", "manhattan", "queens", "staten island"]:
        sn = f"{m.group(0)} {sn}"
    return sn


def short_name(row):
    """Attempts to guess the common "short name" for a school.
For example, the full name might be "P.S. 015 Roberto Clemente".
This school is probably commonly referred to as PS 15."""
    sn = row.school_name.upper()
    if "P.S." in sn or "P. S." in sn:
        return f"PS {row.school_num}"
    if "M.S." in sn or "M. S." in sn:
        return f"MS {row.school_num}"
    if "I. S." in sn or "I.S." in sn:
        return f"IS {row.school_num}"

    return f"{row.school_type} {row.school_num}"


def load_school_demographics():
    """ Loads the NYC school-level demographic data from the
    open data portal and create a dataframe.

    Adds new columns to the data:

           short_name: the best guess for the common name of the school (e.g. PS 9)
                       or "" if none exists
             district: the school district number [1..32, 75, 79, 84]
                 boro: borough code string ["M","B","X","Q","R"]
            boro_name: string of the full borough name
                 year: the academic year as an integer representing the calendar
                       year in the fall of the school year
          white_asian: combined number of white and asian students
            non_white: total number of non white students
      non_white_asian: total number of non white or asian students
       black_hispanic: total number of black and hispanic students
              charter: 1 for charter schools, 0 for community public schools
                 beds: the New York State BEDS code for the school
                  zip: school zip code
         geo_district: the geographic school district where the school is located
                       will differ for schools where district is > 32

    Returns
    --------------
    DataFrame
        a pandas DataFrame holding school demographic data for all of the schools
        in the data portal

    """

    # try to load it locally to save time
    path = config.urls["demographics"].filename
    df = load(path)
    df.zip = df.zip.fillna(0).astype("int32")
    df.beds = df.beds.fillna(0).astype("int64")
    return df

def get_demographics(df):
    """Loads and cleans school demographic data from a pandas DataFrame.
    The initial data is joined with location data to make it
    more easily merged with other data sets.

    Parameters
    -----------
    df : DataFrame, a "raw" DataFrame of school demographic data from NYC Schools or Open Data Portal

    Returns
    -------
    DataFrame
        a pandas DataFrame holding school demographic data for all of the schools
    """

    boros = {"K":"Brooklyn", "X":"Bronx", "M": "Manhattan", "Q": "Queens", "R": "Staten Island"}

    df["ay"] = df["year"].apply(lambda year: int(year.split("-")[0]))
    df["district"] = df["dbn"].apply(lambda dbn: int(dbn[:2]))

    df["boro"] = df["dbn"].apply(lambda dbn: boros[dbn[2]])
    df["school_num"] = df.dbn.apply(lambda dbn: int(dbn[3:]))
    df["charter"] = df.district.apply(lambda x: 1 if x == 84 else 0)

    # figure out what grades they teach
    df["pk"] = df["grade_pk"] > 0
    df["elementary"] = df["grade_2"] > 0
    df["middle"] = df["grade_7"] > 0
    df["hs"] = df["grade_10"] > 0

    # make it easier to look up schools
    df["school_type"] = df.apply(school_type, axis=1)
    df["clean_name"] = df.apply(lambda row: clean_name(row.school_name), axis=1)
    df["short_name"] = df.apply(short_name, axis=1)

    df["poverty"] = df.apply(lambda row: str_count(row, "poverty", "total_enrollment"), axis = 1)
    df["poverty_1"] = df.apply(lambda row: str_pct(row, "poverty_1", "total_enrollment"), axis = 1)
    df["economic_need_index"] = df.apply(lambda row: str_pct(row, "economic_need_index", "total_enrollment"), axis = 1)
    
    df = df.rename(columns=demo.default_map)
    
    # calculate school type based on district number
    df["school_type"] = None
    df.loc[(df.district > 0)  & (df.district < 33), "school_type"] = "community"
    df.loc[df.district == 84, "school_type"] = "charter"
    df.loc[df.district == 75, "school_type"] = "d75"
    df.loc[df.district == 79, "school_type"] = "alternative"
    df = join_loc_data(df)
    df = df[demo.default_cols]


    return df


def get_demo_2006():
    """Load the 2006 demographic data from the NYC Open Data Portal"""
    df = pd.read_csv(config.urls["demographics"].data_urls["2006"])
    def map_col(c):
        if c.startswith("grade"):
            return f"grade_{c[len('grade'):]}"
        if c == "sped_percent":
            return "swd_pct"
        if c == "sped_num":
            return "swd_n"
        if c == "prek":
            return "grade_pk"
        if c == "k":
            return f"grade_{c}"
        if c.endswith("_num"):
            return c.replace("_num", "_n")
        if c.endswith("_percent"):
            return c.replace("_percent", "_pct")
        if c.endswith("_per"):
            return c.replace("_per", "_pct")
        if c == "name":
            return "school_name"
        return c

    df = df.rename(columns=map_col)

    # parse AY and year
    df["ay"] = df.schoolyear // 10000
    df["year"] = df.ay.apply(lambda x: f"{x}-{((x % 10) + 1):02}")

    # make percentages real
    def check(n):
        if pd.isna(n) or pd.isnull(n):
            return 0
        try:
            float(n)
            return float(n)
        except:
            return 0
    # fill missing free lunch data
    df.fl_pct = df.fl_pct.apply(check)
    df.frl_pct = df.frl_pct.apply(check)

    # fix the grades and percentages
    for c in df.columns:
        if c.startswith("grade_"):
            df[c] = pd.to_numeric(df[c], downcast='integer', errors='coerce')
        if c.endswith("_pct"):
            df[c] = np.round(df[c] / 100, 4)

    df["poverty_pct"] = np.maximum(df.fl_pct, df.frl_pct)
    df["poverty_n"] = np.floor(df.poverty_pct * df.total_enrollment).astype(int)

    df["district"] = df["dbn"].apply(lambda dbn: int(dbn[:2]))

    boros = {"K": "Brooklyn", "X": "Bronx", "M": "Manhattan", "Q": "Queens", "R": "Staten Island"}
    df["boro"] = df["dbn"].apply(lambda dbn: boros[dbn[2]])
    df["school_num"] = df.dbn.apply(lambda dbn: int(dbn[3:]))
    df["charter"] = 0


    # figure out what grades they teach
    df["pk"] = df["grade_pk"] > 0
    df["elementary"] = df["grade_2"] > 0
    df["middle"] = df["grade_7"] > 0
    df["hs"] = df["grade_10"] > 0

    # make it easier to look up schools
    df["school_type"] = df.apply(school_type, axis=1)
    df["clean_name"] = df.apply(lambda row: clean_name(row.school_name), axis=1)
    df["short_name"] = df.apply(short_name, axis=1)

    df['eni'] = 0
    df['missing_race_ethnicity_data_n'] = 0
    df['missing_race_ethnicity_data_pct'] = 0
    df['multi_racial_n'] = 0
    df['multi_racial_pct'] = 0
    df['native_american_n'] = 0
    df['native_american_pct'] = 0

    df = join_loc_data(df)
    return df[demo.default_cols]


def get_demo_2013():
    def map_col(c):
        if c.endswith("_1") and not c.startswith("grade"):
            return c.replace("_1", "")
        if c.endswith("_2") and not c.startswith("grade"):
            return c.replace("_2", "_1")
        if c == "grade_pk_half_day_full_day":
            return "grade_3k_pk_half_day_full"
        return c

    df13 = pd.read_csv(config.urls["demographics"].data_urls["2013"])

    df13 = df13.rename(columns=map_col)
    df13["multi_racial"] = df13['multiple_race_categories_not_represented']
    df13["multi_racial_1"] = df13['multiple_race_categories_not_represented_1']
    df13['missing_race_ethnicity_data'] = 0
    df13['missing_race_ethnicity_data_1'] = 0
    df13['native_american'] = 0
    df13['native_american_1'] = 0

    df13.rename(columns={"grade_3k_pk_half_day_full": "grade_pk"}, inplace=True)
    df13["neither_female_nor_male"] = 0
    df13["neither_female_nor_male_pct"] = 0
    df13 = get_demographics(df13)
    # fix the percentages
    for c in df13.columns:
        if c.startswith("grade_"):
            df13[c] = pd.to_numeric(df13[c], downcast='integer', errors='coerce')
        if c.endswith("_pct"):
            df13[c] = np.round(df13[c] / 100, 4)

    # the other data set has 2016 data, don't duplicate it
    return df13[df13.ay.isin([2013, 2014, 2015, 2017])]


def get_demo_2016():
    demo_2016 = pd.read_csv(config.urls["demographics"].data_urls["2016"])
    demo_2016 = demo_2016[demo_2016.year == "2016-17"]
    # make the columns match the most recent data set
    demo_2016.rename(
        columns={"grade_3k_pk_half_day_full": "grade_pk"}, inplace=True)
    demo_2016["neither_female_nor_male"] = 0
    demo_2016["neither_female_nor_male_pct"] = 0
    return get_demographics(demo_2016)

def save_demographics():
    demo_2006 = get_demo_2006()
    demo_2013 = get_demo_2013()
    demo_2016 = get_demo_2016()
    demo_2022 = get_demo_2022()
    # join the dataframes
    df = pd.concat([demo_2006, demo_2013, demo_2016, demo_2022])

    df.to_csv(__demo_filename, index=False)
    return df


def get_demo_2022():
    """Read the latest demographic data from an Excel download"""
    def xls_cols(col):
        d = {
            "multi-racial": "multi_racial",
            "multi-racial_1": "multi_racial_1",
            "grade_pk_(half_day_&_full_day)": "grade_pk",
            "missing_race/ethnicity_data": "missing_race_ethnicity_data",
            "missing_race/ethnicity_data_1": "missing_race_ethnicity_data_1"
        }
        col_name = col.lower().replace(" ", "_")
        if col_name[0] == "%":
            col_name = col_name[2:] + "_1"
        elif col_name[0] == "#":
            col_name = col_name[2:]
        if col_name in d:
            return d[col_name]
        return col_name

    xls = pd.read_excel( config.urls["demographics"].data_urls["2022"], sheet_name=None)
    df = xls["School"]
    df.rename(columns=xls_cols, inplace=True)
    df = get_demographics(df)
    return df

def join_loc_data(df):
    """Join NYS BEDS id, zip code, and other location data.
    """
    school_loc = geo.load_school_locations()
    # with the left join, we might be missing zip and beds for some schools
    df = df.merge(school_loc[["dbn", "beds", "zip", "geo_district"]], on="dbn", how="left")
    # df.beds = df.beds.fillna(0)
    df.geo_district = df.geo_district.fillna(0)
    df.geo_district = df.geo_district.astype(int)
    # df.zip = df.zip.fillna(0).astype("int32")
    df.beds = pd.to_numeric(df.beds , downcast='integer', errors='coerce')
    df.beds = df.beds.fillna(0).astype("int64")

    def get_geo(row):
        d = row.district
        if row.geo_district > 0:
            return row.geo_district
        if d > 0 and d <=32:
            return d
        return 0
    df.geo_district = df.apply(get_geo, axis=1)
    return df

def search(df, qry):
    """Search a DataFrame for a school using fuzzy logic.


    Parameters
    ----------
    df : DataFrame
         a pandas DataFrame containing school data, such as the
         data returned from `load_school_demographics()`
    qry : the school name or search term to look for in the data set

    Returns
    ---------
    DataFrame
        the schools that match the `qry` or an empty DataFrame
        if no matches were found

    """
    t = df.copy(deep=False)
    latest = t.ay.max()
    # just check the lastest year
    t = t[t["ay"] == latest]

    # see if there's an exact name match
    q = clean_name(qry)
    results = t.query(f"clean_name == '{q}'")
    if len(q) > 0 and len(results) > 0:
        return results

    # look it up by number
    results = t.query(f"short_name == '{qry.upper()}'")
    if len(results) > 0:
        return results

    # fuzzy search -- add a fuzzy search match column
    t["match"] = t.school_name.apply(lambda sn: fuzz.token_set_ratio(qry, sn))
    t = t.query("match > 80")
    t = pd.DataFrame(t)
    # now sort the results based on the ratio match
    t["match"] = t.clean_name.apply(lambda sn: fuzz.ratio(qry, sn))
    results = t.sort_values(by=["match"], ascending=False)

    return results


def load_hs_directory(ay=2021):
    """Loads the NYC High School Directory data from the NYC Open Data
    Portal. This is a thin wrapper around `pd.read_csv()` and
    requires an internet connection. By default, the most recent directory
    is returned. For a different academic year, pass the `ay` parameter.
    Data is available for academic years 2013-2021. The data varies greatly
    from year to year, so they are not compiled into a single DataFrame.

    Parameters
    -----------
    ay : int , default 2021
         the academic year for the directory data

    Returns
    --------
    DataFrame
        a pandas DataFrame holding the high school directory data

    """
    urls = config.urls["hs_directory"].data_urls
    ay = str(ay)

    if ay not in urls.keys():
        raise ValueError(f"No data for academic year: {ay}")
    url = urls[ay]
    df = pd.read_csv(url)
    df["ay"] = int(ay)
    return df
